//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2000 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

// MACROS ------------------------------------------------------------------

#define MAX_TAGGED_LINES	64

// Keys

#define	TXT_KEY_STEEL		"STEEL KEY"
#define	TXT_KEY_CAVE		"CAVE KEY"
#define	TXT_KEY_AXE			"AXE KEY"
#define	TXT_KEY_FIRE		"FIRE KEY"
#define	TXT_KEY_EMERALD		"EMERALD KEY"
#define	TXT_KEY_DUNGEON		"DUNGEON KEY"
#define	TXT_KEY_SILVER		"SILVER KEY"
#define	TXT_KEY_RUSTED		"RUSTED KEY"
#define	TXT_KEY_HORN		"HORN KEY"
#define	TXT_KEY_SWAMP		"SWAMP KEY"
#define TXT_KEY_CASTLE		"CASTLE KEY"

// TYPES -------------------------------------------------------------------

struct tagged_line_t
{
	line_t 	*line;
	int 	lineTag;
};
// FUNCTION PROTOTYPES -----------------------------------------------------

//
//	Thing specials
//
boolean EV_ThingProjectile(int *args, boolean gravity);
boolean EV_ThingSpawn(int *args, boolean fog);
boolean EV_ThingActivate(int tid);
boolean EV_ThingDeactivate(int tid);
boolean EV_ThingRemove(int tid);
boolean EV_ThingDestroy(int tid);
//
//	Teleportation
//
boolean EV_Teleport(int tid, mobj_t *thing, boolean fog);
//
//	Quake
//
boolean A_LocalQuake(int *args, mobj_t *actor);
//
//
//
void P_ThrustMobj(mobj_t *mo, angle_t angle, float move);
boolean EV_LineSearchForPuzzleItem(line_t *line, int *args, mobj_t *mo);

// DATA DECLARATIONS -------------------------------------------------------

tagged_line_t	TaggedLines[MAX_TAGGED_LINES];
int 			TaggedLineCount;

string TextKeyMessages[] =
{
	TXT_KEY_STEEL,
	TXT_KEY_CAVE,
	TXT_KEY_AXE,
	TXT_KEY_FIRE,
	TXT_KEY_EMERALD,
	TXT_KEY_DUNGEON,
	TXT_KEY_SILVER,
	TXT_KEY_RUSTED,
	TXT_KEY_HORN,
	TXT_KEY_SWAMP,
	TXT_KEY_CASTLE
};

// CODE --------------------------------------------------------------------

//============================================================================
//
// CheckedLockedDoor
//
//============================================================================

boolean CheckedLockedDoor(mobj_t *mo, int lock)
{
	if (!mo->player)
	{
		return false;
	}
	if (!lock)
	{	
		return true;
	}
	if (!(mo->player->keys & (1 << (lock - 1))))
	{
		centerprint(mo->player, "YOU NEED THE %s\n", TextKeyMessages[lock - 1]);
		StartSoundName(mo, "DoorLocked");
		return false;
	}
	return true;
}

//==========================================================================
//
//  StartLockedACS
//
//==========================================================================

boolean StartLockedACS(line_t *line, int *args, mobj_t *mo, int side)
{
	int i;
	int newArgs[5];

	if (!CheckedLockedDoor(mo, args[4]))
	{
		return false;
	}
	for (i = 0; i < 4; i++)
	{
		newArgs[i] = args[i];
	}
	newArgs[4] = 0;
	return StartACS(newArgs[0], newArgs[1], &newArgs[2], mo, line, side);
}

//=========================================================================
//
// EV_SectorSoundChange
//
//=========================================================================

boolean EV_SectorSoundChange(int *args)
{
	int 	secNum;
	boolean rtn;

	if (!args[0])
	{
		return false;
	}
	secNum = -1;
	rtn = false;
	while ((secNum = FindSectorFromTag(args[0], secNum)) >= 0)
	{
		level->sectors[secNum].seqType = args[1];
		rtn = true;
	}
	return rtn;
}

//==========================================================================
//
//  ExecuteLineSpecial
//
//==========================================================================

boolean ExecuteLineSpecial(int special, int *args, line_t *line, int side,
	mobj_t *mo)
{
	boolean buttonSuccess;

	buttonSuccess = false;
	switch (special)
	{
		case 1: // Poly Start Line
			break;
		case 2: // Poly Rotate Left
			buttonSuccess = EV_RotatePoly(line, args, 1, false);
			break;
		case 3: // Poly Rotate Right
			buttonSuccess = EV_RotatePoly(line, args, -1, false);
			break;
		case 4: // Poly Move
			buttonSuccess = EV_MovePoly(line, args, false, false);
			break;
		case 5: // Poly Explicit Line:  Only used in initialization
			break;
		case 6: // Poly Move Times 8
			buttonSuccess = EV_MovePoly(line, args, true, false);
			break;
		case 7: // Poly Door Swing
			buttonSuccess = EV_OpenPolyDoor(line, args, PODOOR_SWING);
			break;
		case 8: // Poly Door Slide
			buttonSuccess = EV_OpenPolyDoor(line, args, PODOOR_SLIDE);
			break;
		case 10: // Door Close
			buttonSuccess = EV_DoDoor(line, args, DREV_CLOSE);
			break;
		case 11: // Door Open
			if(!args[0])
			{
				buttonSuccess = EV_VerticalDoor(line, mo);
			}
			else
			{
				buttonSuccess = EV_DoDoor(line, args, DREV_OPEN);
			}
			break;
		case 12: // Door Raise
			if(!args[0])
			{
				buttonSuccess = EV_VerticalDoor(line, mo);
			}
			else
			{
				buttonSuccess = EV_DoDoor(line, args, DREV_NORMAL);
			}
			break;
		case 13: // Door Locked_Raise
			if (CheckedLockedDoor(mo, args[3]))
			{
				if(!args[0])
				{
					buttonSuccess = EV_VerticalDoor(line, mo);
				}
				else
				{
					buttonSuccess = EV_DoDoor(line, args, DREV_NORMAL);
				}
			}
			break;
		case 20: // Floor Lower by Value
			buttonSuccess = EV_DoFloor(line, args, FLEV_LOWERFLOORBYVALUE);
			break;
		case 21: // Floor Lower to Lowest
			buttonSuccess = EV_DoFloor(line, args, FLEV_LOWERFLOORTOLOWEST);
			break;
		case 22: // Floor Lower to Nearest
			buttonSuccess = EV_DoFloor(line, args, FLEV_LOWERFLOOR);
			break;
		case 23: // Floor Raise by Value
			buttonSuccess = EV_DoFloor(line, args, FLEV_RAISEFLOORBYVALUE);
			break;
		case 24: // Floor Raise to Highest
			buttonSuccess = EV_DoFloor(line, args, FLEV_RAISEFLOOR);
			break;
		case 25: // Floor Raise to Nearest
			buttonSuccess = EV_DoFloor(line, args, FLEV_RAISEFLOORTONEAREST);
			break;
		case 26: // Stairs Build Down Normal
			buttonSuccess = EV_BuildStairs(line, args, -1, STAIRS_NORMAL);
			break;
		case 27: // Build Stairs Up Normal
			buttonSuccess = EV_BuildStairs(line, args, 1, STAIRS_NORMAL);
			break;
		case 28: // Floor Raise and Crush
			buttonSuccess = EV_DoFloor(line, args, FLEV_RAISEFLOORCRUSH);
			break;
		case 29: // Build Pillar (no crushing)
			buttonSuccess = EV_BuildPillar(line, args, false);
			break;
		case 30: // Open Pillar
			buttonSuccess = EV_OpenPillar(line, args);
			break;
		case 31: // Stairs Build Down Sync
			buttonSuccess = EV_BuildStairs(line, args, -1, STAIRS_SYNC);
			break;
		case 32: // Build Stairs Up Sync
			buttonSuccess = EV_BuildStairs(line, args, 1, STAIRS_SYNC);
			break;
		case 35: // Raise Floor by Value Times 8
			buttonSuccess = EV_DoFloor(line, args, FLEV_RAISEBYVALUETIMES8);
			break;
		case 36: // Lower Floor by Value Times 8
			buttonSuccess = EV_DoFloor(line, args, FLEV_LOWERBYVALUETIMES8);
			break;
		case 40: // Ceiling Lower by Value
			buttonSuccess = EV_DoCeiling(line, args, CLEV_LOWERBYVALUE);
			break;
		case 41: // Ceiling Raise by Value
			buttonSuccess = EV_DoCeiling(line, args, CLEV_RAISEBYVALUE);
			break;
		case 42: // Ceiling Crush and Raise
			buttonSuccess = EV_DoCeiling(line, args, CLEV_CRUSHANDRAISE);
			break;
		case 43: // Ceiling Lower and Crush
			buttonSuccess = EV_DoCeiling(line, args, CLEV_LOWERANDCRUSH);
			break;
		case 44: // Ceiling Crush Stop
			buttonSuccess = EV_CeilingCrushStop(line, args);
			break;
		case 45: // Ceiling Crush Raise and Stay
			buttonSuccess = EV_DoCeiling(line, args, CLEV_CRUSHRAISEANDSTAY);
			break;
		case 46: // Floor Crush Stop
			buttonSuccess = EV_FloorCrushStop(line, args);
			break;
		case 60: // Plat Perpetual Raise
			buttonSuccess = EV_DoPlat(line, args, PLAT_PERPETUALRAISE);
			break;
		case 61: // Plat Stop
			EV_StopPlat(line, args);
			break;
		case 62: // Plat Down-Wait-Up-Stay
			buttonSuccess = EV_DoPlat(line, args, PLAT_DOWNWAITUPSTAY);
			break;
		case 63: // Plat Down-by-Value*8-Wait-Up-Stay
			buttonSuccess = EV_DoPlat(line, args, PLAT_DOWNBYVALUEWAITUPSTAY);
			break;
		case 64: // Plat Up-Wait-Down-Stay
			buttonSuccess = EV_DoPlat(line, args, PLAT_UPWAITDOWNSTAY);
			break;
		case 65: // Plat Up-by-Value*8-Wait-Down-Stay
			buttonSuccess = EV_DoPlat(line, args, PLAT_UPBYVALUEWAITDOWNSTAY);
			break;
		case 66: // Floor Lower Instant * 8
			buttonSuccess = EV_DoFloor(line, args, FLEV_LOWERTIMES8INSTANT);
			break;
		case 67: // Floor Raise Instant * 8
			buttonSuccess = EV_DoFloor(line, args, FLEV_RAISETIMES8INSTANT);
			break;
		case 68: // Floor Move to Value * 8
			buttonSuccess = EV_DoFloor(line, args, FLEV_MOVETOVALUETIMES8);
			break;
		case 69: // Ceiling Move to Value * 8
			buttonSuccess = EV_DoCeiling(line, args, CLEV_MOVETOVALUETIMES8);
			break;
		case 70: // Teleport
			if (side == 0)
			{ // Only teleport when crossing the front side of a line
				buttonSuccess = EV_Teleport(args[0], mo, true);
			}
			break;
		case 71: // Teleport, no fog
			if(side == 0)
			{ // Only teleport when crossing the front side of a line
				buttonSuccess = EV_Teleport(args[0], mo, false);
			}
			break;
		case 72: // Thrust Mobj
			if (!side) // Only thrust on side 0
			{
				P_ThrustMobj(mo, args[0] * (ANG90 / 64), itof(args[1]));
				buttonSuccess = 1;
			}
			break;
		case 73: // Damage Mobj
			if(args[0])
			{
				DamageMobj(mo, NULL, NULL, args[0]);
			}
			else
			{ // If arg1 is zero, then guarantee a kill
				DamageMobj(mo, NULL, NULL, 10000);
			}
			buttonSuccess = 1;
			break;
		case 74: // Teleport_NewMap
			if(side == 0)
			{ // Only teleport when crossing the front side of a line
				if(!(mo && mo->player && mo->player->playerstate
					== PST_DEAD)) // Players must be alive to teleport
				{
					G_Completed(args[0], args[1]);
					buttonSuccess = true;
				}
			}
			break;
		case 75: // Teleport_EndGame
			if(side == 0)
			{ // Only teleport when crossing the front side of a line
				if(!(mo && mo->player && mo->player->playerstate
					== PST_DEAD)) // Players must be alive to teleport
				{
					buttonSuccess = true;
					if(deathmatch)
					{ // Winning in deathmatch just goes back to map 1
						G_Completed(1, 0);
					}
					else
					{ // Passing -1, -1 to G_Completed() starts the Finale
						G_Completed(-1, -1);
					}
				}
			}
			break;
		case 80: // ACS_Execute
			buttonSuccess =
				StartACS(args[0], args[1], &args[2], mo, line, side);
			break;
		case 81: // ACS_Suspend
			buttonSuccess = SuspendACS(args[0], args[1]);
			break;
		case 82: // ACS_Terminate
			buttonSuccess = TerminateACS(args[0], args[1]);
			break;
		case 83: // ACS_LockedExecute
			buttonSuccess = StartLockedACS(line, args, mo, side);
			break;
		case 90: // Poly Rotate Left Override
			buttonSuccess = EV_RotatePoly(line, args, 1, true);
			break;
		case 91: // Poly Rotate Right Override
			buttonSuccess = EV_RotatePoly(line, args, -1, true);
			break;
		case 92: // Poly Move Override
			buttonSuccess = EV_MovePoly(line, args, false, true);
			break;
		case 93: // Poly Move Times 8 Override
			buttonSuccess = EV_MovePoly(line, args, true, true);
			break;
		case 94: // Build Pillar Crush 
			buttonSuccess = EV_BuildPillar(line, args, true);
			break;
		case 95: // Lower Floor and Ceiling
			buttonSuccess = EV_DoFloorAndCeiling(line, args, false);
			break;
		case 96: // Raise Floor and Ceiling
			buttonSuccess = EV_DoFloorAndCeiling(line, args, true);
			break;
		case 109: // Force Lightning
			buttonSuccess = true;
			P_ForceLightning();
			break;
		case 110: // Light Raise by Value
			buttonSuccess = EV_SpawnLight(line, args, LITE_RAISEBYVALUE);
			break; 
		case 111: // Light Lower by Value
			buttonSuccess = EV_SpawnLight(line, args, LITE_LOWERBYVALUE);
			break; 
		case 112: // Light Change to Value
			buttonSuccess = EV_SpawnLight(line, args, LITE_CHANGETOVALUE);
			break; 
		case 113: // Light Fade
			buttonSuccess = EV_SpawnLight(line, args, LITE_FADE);
			break; 
		case 114: // Light Glow
			buttonSuccess = EV_SpawnLight(line, args, LITE_GLOW);
			break; 
		case 115: // Light Flicker
			buttonSuccess = EV_SpawnLight(line, args, LITE_FLICKER);
			break; 
		case 116: // Light Strobe
			buttonSuccess = EV_SpawnLight(line, args, LITE_STROBE);
			break; 
		case 120: // Quake Tremor
			buttonSuccess = A_LocalQuake(args, mo);
			break;
		case 129: // UsePuzzleItem
			buttonSuccess = EV_LineSearchForPuzzleItem(line, args, mo);
			break;
		case 130: // Thing_Activate
			buttonSuccess = EV_ThingActivate(args[0]);
			break;
		case 131: // Thing_Deactivate
			buttonSuccess = EV_ThingDeactivate(args[0]);
			break;
		case 132: // Thing_Remove
			buttonSuccess = EV_ThingRemove(args[0]);
			break;
		case 133: // Thing_Destroy
			buttonSuccess = EV_ThingDestroy(args[0]);
			break;
		case 134: // Thing_Projectile
			buttonSuccess = EV_ThingProjectile(args, 0);
			break;
		case 135: // Thing_Spawn
			buttonSuccess = EV_ThingSpawn(args, 1);
			break;
		case 136: // Thing_ProjectileGravity
			buttonSuccess = EV_ThingProjectile(args, 1);
			break;
		case 137: // Thing_SpawnNoFog
			buttonSuccess = EV_ThingSpawn(args, 0);
			break;
		case 138: // Floor_Waggle
			buttonSuccess = EV_StartFloorWaggle(args[0], args[1],
				args[2], args[3], args[4]);
			break;
		case 140: // Sector_SoundChange
			buttonSuccess = EV_SectorSoundChange(args);
			break;

		// Line specials only processed during level initialization
		// 100: Scroll_Texture_Left
		// 101: Scroll_Texture_Right
		// 102: Scroll_Texture_Up
		// 103: Scroll_Texture_Down
		// 121: Line_SetIdentification

		// Inert Line specials
		default:
			break;
	}
	return buttonSuccess;
}

//============================================================================
//
//	P_ActivateLine
//
//============================================================================

boolean P_ActivateLine(line_t *line, mobj_t *mo, int side, int activationType)
{
	int			lineActivation;
	boolean 	repeat;
	boolean 	buttonSuccess;

	lineActivation = GET_SPAC(line->flags);
	if (lineActivation != activationType)
	{
		return false;
	}
	if (!mo->player && !(mo->flags & MF_MISSILE))
	{
		if (lineActivation != SPAC_MCROSS)
		{
			// currently, monsters can only activate the MCROSS activation type
 			return false;
		}
		if (line->flags & ML_SECRET)
			return false;           // never open secret doors
	}
	repeat = line->flags & ML_REPEAT_SPECIAL;
	buttonSuccess = false;

	buttonSuccess = ExecuteLineSpecial(line->special, &line->arg1, line,
		side, mo);
	if (!repeat && buttonSuccess)
	{
		// clear the special on non-retriggerable lines
		line->special = 0;
	}
	if ((lineActivation == SPAC_USE || lineActivation == SPAC_IMPACT)
		&& buttonSuccess)
	{
		P_ChangeSwitchTexture(line, repeat);
	}
	return true;
}

//============================================================================
//
//  P_SpawnSpecials
//
//============================================================================

void P_SpawnSpecials(boolean spawn_thinkers)
{
	sector_t        *sector;
	int             i;

	if (spawn_thinkers)
    {
		//
		//      Init special SECTORs
		//
		for (i=0 ; i<level->numsectors ; i++)
		{
			sector = &level->sectors[i];
			if (!sector->special)
				continue;
			switch (sector->special)
			{
			 case 1: // Phased light
				// Hardcoded base, use sector->lightlevel as the index
				P_SpawnPhasedLight(sector, 80, -1);
				break;
			 case 2: // Phased light sequence start
				P_SpawnLightSequence(sector, 1.0);
				break;
			// Specials 3 & 4 are used by the phased light sequences
			 case 201: case 202: case 203: // Scroll_North_xxx
                SpawnScrollingFloor(sector, 0, 1, sector->special - 201);
				break;
			 case 204: case 205: case 206: // Scroll_East_xxx
            	SpawnScrollingFloor(sector, -1, 0, sector->special - 204);
				break;
			 case 207: case 208: case 209: // Scroll_South_xxx
                SpawnScrollingFloor(sector, 0, -1, sector->special - 207);
				break;
			 case 210: case 211: case 212: // Scroll_West_xxx
                SpawnScrollingFloor(sector, 1, 0, sector->special - 210);
				break;
			 case 213: case 214: case 215: // Scroll_NorthWest_xxx
                SpawnScrollingFloor(sector, 1, 1, sector->special - 213);
				break;
			 case 216: case 217: case 218: // Scroll_NorthEast_xxx
                SpawnScrollingFloor(sector, -1, 1, sector->special - 216);
				break;
			 case 219: case 220: case 221: // Scroll_SouthEast_xxx
                SpawnScrollingFloor(sector, -1, -1, sector->special - 219);
				break;
			 case 222: case 223: case 224: // Scroll_SouthWest_xxx
                SpawnScrollingFloor(sector, 1, -1, sector->special - 222);
				break;
			}
		}

		//
		//      Init line EFFECTs
		//
		for (i = 0; i < level->numlines; i++)
		{
			switch (level->lines[i].special)
			{
			 case 100: // Scroll_Texture_Left
				SpawnWallScroller(&level->lines[i], 1, 0);
				break;
			 case 101: // Scroll_Texture_Right
				SpawnWallScroller(&level->lines[i], -1, 0);
				break;
			 case 102: // Scroll_Texture_Up
				SpawnWallScroller(&level->lines[i], 0, 1);
				break;
			 case 103: // Scroll_Texture_Down
				SpawnWallScroller(&level->lines[i], 0, -1);
				break;
			}
		}
	}

	TaggedLineCount = 0;
	for (i = 0; i < level->numlines; i++)
	{
		switch (level->lines[i].special)
		{
			case 121: // Line_SetIdentification
				if (level->lines[i].arg1)
				{
					if (TaggedLineCount == MAX_TAGGED_LINES)
					{
						Error("P_SpawnSpecials: MAX_TAGGED_LINES exceeded.");
					}
					TaggedLines[TaggedLineCount].line = &level->lines[i];
					TaggedLines[TaggedLineCount++].lineTag = level->lines[i].arg1;
				}
				level->lines[i].special = 0;
				break;
		}
	}

	activeceilings.next = activeceilings.prev = &activeceilings;
	activeplats.next = activeplats.prev = &activeplats;
	activefloors.next = activefloors.prev = &activefloors;
}

//==========================================================================
//
//  P_UpdateSpecials
//
//==========================================================================

void P_UpdateSpecials(void)
{
}

//==========================================================================
//
//	P_FindLine
//
//==========================================================================

line_t *P_FindLine(int lineTag, int *searchPosition)
{
	int i;

	for (i = *searchPosition + 1; i < TaggedLineCount; i++)
	{
		if (TaggedLines[i].lineTag == lineTag)
		{
			*searchPosition = i;
			return TaggedLines[i].line;
		}
	}
	*searchPosition = -1;
	return NULL;
}

