//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2000 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

#define	BODYQUESIZE				32

mobj_t*	bodyque[BODYQUESIZE];
int		bodyqueslot;

//==========================================================================
//
//  SetPlayerRunState
//
//==========================================================================

void SetPlayerRunState(player_t* player)
{
	if (player->mo->statenum == S_PLAY)
	{
		SetMobjState(player->mo, S_PLAY_RUN1);
    }
}

//==========================================================================
//
//	G_CheckSpot
//
//	Returns false if the player cannot be respawned at the given mthing_t
// spot because something is occupying it
//
//==========================================================================

boolean G_CheckSpot(int playernum, mthing_t* mthing)
{
    float			x;
    float			y;
    subsector_t*	ss; 
    angle_t			an;
    mobj_t*			mo;
	int				mtnum;
    int				sfxnum;
    int				i;
	
    if (!players[playernum].mo)
    {
		// first spawn of level, before corpses
		for (i=0 ; i<playernum ; i++)
	    	if (players[i].mo->origin.x == mthing->x &&
	    		players[i].mo->origin.y == mthing->y)
				return false;
		return true;
    }

    x = mthing->x;
    y = mthing->y;
	 
    if (!P_CheckPosition(players[playernum].mo, x, y))
	{
		return false;
	}

    // flush an old corpse if needed 
    if (bodyqueslot >= BODYQUESIZE) 
		P_RemoveMobj(bodyque[bodyqueslot % BODYQUESIZE]);
    bodyque[bodyqueslot % BODYQUESIZE] = players[playernum].mo;
    bodyqueslot++; 

    // spawn a teleport fog 
    ss = PointInSubsector(x, y);
    an = ANG45 * (mthing->angle / 45);
 
	mo = P_SpawnMobj(x + 20.0 * cos(an), y + 20.0 * sin(an),
		ss->sector->floorheight, MT_TFOG);
	StartSoundName(mo, "Teleport");	// don't start sound on first frame

    return true; 
}

//==========================================================================
//
//	P_SpawnPlayer
//
//	Called when a player is spawned on the level. Most of the player
// structure stays unchanged between levels.
//
//==========================================================================

void P_SpawnPlayer(mthing_t* mthing)
{
    player_t*	p;
    float		x;
    float		y;
    float		z;

    mobj_t*		mobj;

    p = &players[mthing->type - 1];

    // not playing?
    if (!p->active)
		return;
		
    if (p->playerstate == PST_REBORN)
		G_PlayerReborn(mthing->type - 1);

    x = mthing->x;
    y = mthing->y;
    z = ONFLOORZ;
    mobj = P_SpawnMobj(x, y, z, MT_PLAYER);

    // set color translations for player sprites
	mobj->translation = p->color;
		
    mobj->angles.yaw = ANG45 * (mthing->angle / 45);
    mobj->player = p;
    mobj->health = p->health;

    p->mo = mobj;
    p->playerstate = PST_LIVE;	
    p->refire = 0;
    p->damagecount = 0;
    p->bonuscount = 0;
    p->extralight = 0;
    p->fixedcolormap = 0;
    p->viewheight = VIEWHEIGHT;

    // setup gun psprite
    SetupPsprites(p);
    
    // give all cards in death match mode
    if (deathmatch)
    {
    	p->keys = 0x3f;
	}
			
	// wake up the status bar
	SB_Start();
}

//==========================================================================
//
//  SpawnPlayerMissile
//
//	Tries to aim at a nearby monster
//
//==========================================================================

mobj_t* SpawnPlayerMissile(mobj_t* source, int type)
{
    mobj_t		*th;
    TAVec		angles;
	TVec		vforward;
	float		x;
	float		y;
	float		z;
    
    // see which target is to be aimed at
    avcpy(source->angles, angles);
	AimLineAttack(source, &angles, 16.0 * 64.0);
    if (!linetarget)
   	{
		angles.yaw += 1 << 26;
		AimLineAttack(source, &angles, 16.0 * 64.0);
		if (!linetarget)
		{
    		angles.yaw -= 2 << 26;
			AimLineAttack(source, &angles, 16.0 * 64.0);
			if (!linetarget)
			{
				angles.yaw += 1 << 26;
			}
		}
    }

    x = source->origin.x;
    y = source->origin.y;
    z = source->origin.z + 32.0;
	
    th = P_SpawnMobj(x, y, z, type);

    if (th->sound_sight)
		StartSound(th, th->sound_sight);

    th->target = source;
    avcpy(angles, th->angles);
	AngleVector(&angles, &vforward);
    vscale(vforward, th->speed, th->velocity);

    CheckMissileSpawn(th);

	return th;
}


