//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "d3d_local.h"

// MACROS ------------------------------------------------------------------

#define MTEX_VERTEX_F	D3DFVF_XYZ | D3DFVF_TEX2
#define PART_VERTEX_F	D3DFVF_XYZ | D3DFVF_DIFFUSE

// TYPES -------------------------------------------------------------------

struct MTEX_VERTEX
{
    float		x;			// Homogeneous coordinates
    float		y;
    float		z;
    float		texs;		// Texture coordinates
    float		text;
    float		lights;		// Lightmap coordinates
    float		lightt;
};

struct PART_VERTEX
{
	TVec		origin;
	dword		color;
};

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

//==========================================================================
//
//	TDirect3DDrawer::FlushCaches
//
//==========================================================================

void TDirect3DDrawer::FlushCaches(bool free_blocks)
{
	int				i;
	surfcache_t		*blines;
	surfcache_t		*block;

	if (free_blocks)
	{
		for (i = 0; i < NUM_BLOCK_SURFS; i++)
		{
			for (blines = cacheblocks[i]; blines; blines = blines->bnext)
			{
				for (block = blines; block; block = block->lnext)
				{
					if (block->owner)
						*block->owner = NULL;
				}
			}
		}
	}

	memset(blockbuf, 0, sizeof(blockbuf));
	freeblocks = NULL;
	for (i = 0; i < NUM_CACHE_BLOCKS; i++)
	{
		blockbuf[i].chain = freeblocks;
		freeblocks = &blockbuf[i];
	}
	for (i = 0; i < NUM_BLOCK_SURFS; i++)
	{
		cacheblocks[i] = freeblocks;
		freeblocks = freeblocks->chain;
		cacheblocks[i]->width = BLOCK_WIDTH;
		cacheblocks[i]->height = BLOCK_HEIGHT;
		cacheblocks[i]->blocknum = i;
	}
}

//==========================================================================
//
//	TDirect3DDrawer::AllocBlock
//
//==========================================================================

surfcache_t	*TDirect3DDrawer::AllocBlock(int width, int height)
{
	int				i;
	surfcache_t		*blines;
	surfcache_t		*block;
	surfcache_t		*other;

	for (i = 0; i < NUM_BLOCK_SURFS; i++)
	{
		for (blines = cacheblocks[i]; blines; blines = blines->bnext)
		{
			if (blines->height != height)
			{
				continue;
			}
			for (block = blines; block; block = block->lnext)
			{
				if (block->owner)
					continue;
				if (block->width < width)
					continue;
				if (block->width > width)
				{
					if (!freeblocks)
						Sys_Error("No more free blocks");
					other = freeblocks;
					freeblocks = other->chain;
					other->s = block->s + width;
					other->t = block->t;
					other->width = block->width - width;
					other->height = block->height;
					other->lnext = block->lnext;
					if (other->lnext)
						other->lnext->lprev = other;
					block->lnext = other;
					other->lprev = block;
					block->width = width;
					other->owner = NULL;
					other->blocknum = i;
				}
				return block;
			}
		}
	}

	for (i = 0; i < NUM_BLOCK_SURFS; i++)
	{
		for (blines = cacheblocks[i]; blines; blines = blines->bnext)
		{
			if (blines->height < height)
			{
				continue;
			}
			if (blines->lnext)
			{
				continue;
			}

			block = blines;
			if (block->height > height)
			{
				if (!freeblocks)
					Sys_Error("No more free blocks");
				other = freeblocks;
				freeblocks = other->chain;
				other->s = 0;
				other->t = block->t + height;
				other->width = block->width;
				other->height = block->height - height;
				other->lnext = NULL;
				other->lprev = NULL;
				other->bnext = block->bnext;
				if (other->bnext)
					other->bnext->bprev = other;
				block->bnext = other;
				other->bprev = block;
				block->height = height;
				other->owner = NULL;
				other->blocknum = i;
			}

			if (!freeblocks)
				Sys_Error("No more free blocks");
			other = freeblocks;
			freeblocks = other->chain;
			other->s = block->s + width;
			other->t = block->t;
			other->width = block->width - width;
			other->height = block->height;
			other->lnext = NULL;
			block->lnext = other;
			other->lprev = block;
			block->width = width;
			other->owner = NULL;
			other->blocknum = i;

			return block;
		}
	}

	Sys_Error("overflow");
	return NULL;
}

//==========================================================================
//
//	TDirect3DDrawer::FreeSurfCache
//
//==========================================================================

void TDirect3DDrawer::FreeSurfCache(surfcache_t *block)
{
	surfcache_t		*other;

	*block->owner = NULL;
	block->owner = NULL;
	if (block->lnext && !block->lnext->owner)
	{
		other = block->lnext;
		block->width += other->width;
		block->lnext = other->lnext;
		if (block->lnext)
			block->lnext->lprev = block;
		other->chain = freeblocks;
		freeblocks = other;
	}
	if (block->lprev && !block->lprev->owner)
	{
		other = block;
		block = block->lprev;
		block->width += other->width;
		block->lnext = other->lnext;
		if (block->lnext)
			block->lnext->lprev = block;
		other->chain = freeblocks;
		freeblocks = other;
	}

	if (block->lprev || block->lnext)
	{
		return;
	}

	if (block->bnext && !block->bnext->lnext)
	{
		other = block->bnext;
		block->height += other->height;
		block->bnext = other->bnext;
		if (block->bnext)
			block->bnext->bprev = block;
		other->chain = freeblocks;
		freeblocks = other;
	}
	if (block->bprev && !block->bprev->lnext)
	{
		other = block;
		block = block->bprev;
		block->height += other->height;
		block->bnext = other->bnext;
		if (block->bnext)
			block->bnext->bprev = block;
		other->chain = freeblocks;
		freeblocks = other;
	}
}

//==========================================================================
//
//	TDirect3DDrawer::CacheSurface
//
//==========================================================================

void TDirect3DDrawer::CacheSurface(surface_t *surface)
{
	surfcache_t     *cache;
	int				smax, tmax;
	int				i, j, bnum;

	//
	// see if the cache holds apropriate data
	//
	cache = surface->cachespots[0];

	if (cache && !cache->dlight && surface->dlightframe != r_dlightframecount
			&& cache->lightlevel == surface->lightlevel)
	{
		bnum = cache->blocknum;
		cache->chain = light_chain[bnum];
		light_chain[bnum] = cache;
		return;
	}

	//
	// determine shape of surface
	//
	smax = (surface->extents[0] >> 4) + 1;
	tmax = (surface->extents[1] >> 4) + 1;
	
	//
	// allocate memory if needed
	//
	if (!cache)     // if a texture just animated, don't reallocate it
	{
		cache = AllocBlock(smax, tmax);
		surface->cachespots[0] = cache;
		cache->owner = &surface->cachespots[0];
		cache->surf = surface;
	}
	
	if (surface->dlightframe == r_dlightframecount)
		cache->dlight = 1;
	else
		cache->dlight = 0;
	cache->lightlevel = surface->lightlevel;

	// calculate the lightings
	R_BuildLightMap(surface, 0);
	bnum = cache->blocknum;
	block_changed[bnum] = true;

	for (j = 0; j < tmax; j++)
	{
		for (i = 0; i < smax; i++)
		{
			light_block[bnum][(j + cache->t) * BLOCK_WIDTH + i + cache->s] = MakeCol16(
				255 - (blocklightsr[j * smax + i] >> 8),
				255 - (blocklightsg[j * smax + i] >> 8),
				255 - (blocklightsb[j * smax + i] >> 8));
		}
	}
	cache->chain = light_chain[bnum];
	light_chain[bnum] = cache;
}

//==========================================================================
//
//	TDirect3DDrawer::DrawPolygon
//
//==========================================================================

void TDirect3DDrawer::DrawPolygon(TVec *cv, int count, int texture, int)
{
	D3DLVERTEX		out[256];
	int				i, j, l;
	float			s, t;
	bool			lightmaped;
	surface_t		*surf = r_surface;

	lightmaped = surf->lightmap != NULL ||
		surf->dlightframe == r_dlightframecount;

	if (lightmaped)
	{
		CacheSurface(surf);
		if (maxMultiTex >= 2)
		{
			return;
		}
		l = 0xffffffff;
	}
	else
	{
		l =	0xff000000 | (surf->lightlevel << 16) |
			(surf->lightlevel << 8) | surf->lightlevel;
	}

	if (texture & TEXF_FLAT)
	{
		SetFlat(texture);
	}
	else
	{
		SetTexture(texture);
	}

	for (i = 0; i < count; i++)
	{
		TVec texpt = cv[i] - r_texorg;
		out[i] = D3DLVERTEX(D3DVECTOR(cv[i].x, cv[i].y, cv[i].z), l, 0,
			DotProduct(texpt, r_saxis) * tex_iw,
			DotProduct(texpt, r_taxis) * tex_ih);
		if (texAdjust)
		{
			out[i].tu -= 0.5 * tex_iw;
			out[i].tv -= 0.5 * tex_ih;
		}
	}

	RenderDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, D3DFVF_LVERTEX, out, count, 0);
}

//==========================================================================
//
//	TDirect3DDrawer::WorldDrawing
//
//==========================================================================

void TDirect3DDrawer::WorldDrawing(void)
{
	D3DLVERTEX		out[256];
	MTEX_VERTEX		mtv[256];
	int				lb, i;
	surfcache_t		*cache;
	float			s, t, lights, lightt;
	surface_t		*surf;
	texinfo_t		*tex;
	word			*buf;

	if (maxMultiTex >= 2)
	{
		RenderDevice->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
		RenderDevice->SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_MODULATE);
		RenderDevice->SetTextureStageState(1, D3DTSS_COLORARG1, D3DTA_TEXTURE);
		RenderDevice->SetTextureStageState(1, D3DTSS_COLORARG2, D3DTA_CURRENT);
		RenderDevice->SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, 1);
		RenderDevice->SetTextureStageState(1, D3DTSS_MAGFILTER, D3DTFG_LINEAR);
		RenderDevice->SetTextureStageState(1, D3DTSS_MINFILTER, D3DTFN_LINEAR);

		for (lb = 0; lb < NUM_BLOCK_SURFS; lb++)
		{
			if (!light_chain[lb])
			{
				continue;
			}

			if (!light_surf[lb])
			{
				light_surf[lb] = CreateSurface(BLOCK_WIDTH, BLOCK_HEIGHT);
				block_changed[lb] = true;
			}
			if (block_changed[lb])
			{
				block_changed[lb] = false;
				buf = LockSurface(light_surf[lb]);
				memcpy(buf, light_block[lb], BLOCK_WIDTH * BLOCK_HEIGHT * 2);
				light_surf[lb]->Unlock(NULL);
			}

			RenderDevice->SetTexture(1, light_surf[lb]);

			for (cache = light_chain[lb]; cache; cache = cache->chain)
			{
				surf = cache->surf;
				tex = surf->texinfo;
				if (tex->pic & TEXF_FLAT)
				{
					SetFlat(tex->pic);
				}
				else
				{
					SetTexture(tex->pic);
				}
				for (i = 0; i < surf->count; i++)
				{
					TVec texpt = surf->verts[i] - tex->texorg;
					s = DotProduct(texpt, tex->saxis);
					t = DotProduct(texpt, tex->taxis);
					lights = (s - surf->texturemins[0]) / 16 + cache->s + 0.5;
					lightt = (t - surf->texturemins[1]) / 16 + cache->t + 0.5;
					if (texAdjust)
					{
						s -= 0.5;
						t -= 0.5;
					}
					if (adjust_linear)
					{
						lights -= 0.5;
						lightt -= 0.5;
					}
					mtv[i].x = surf->verts[i].x;
					mtv[i].y = surf->verts[i].y;
					mtv[i].z = surf->verts[i].z;
					mtv[i].texs = s * tex_iw;
					mtv[i].text = t * tex_ih;
					mtv[i].lights = lights / BLOCK_WIDTH;
					mtv[i].lightt = lightt / BLOCK_HEIGHT;
				}
				RenderDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, MTEX_VERTEX_F, mtv, surf->count, 0);
			}
		}

		RenderDevice->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE);
		RenderDevice->SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_DISABLE);
	}
	else
	{
		RenderDevice->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTFG_LINEAR);
		RenderDevice->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTFN_LINEAR);
		RenderDevice->SetRenderState(D3DRENDERSTATE_SRCBLEND, D3DBLEND_ZERO);
		RenderDevice->SetRenderState(D3DRENDERSTATE_DESTBLEND, D3DBLEND_SRCCOLOR);
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);
		RenderDevice->SetRenderState(D3DRENDERSTATE_ZWRITEENABLE, FALSE);	// don't bother writing Z

		for (lb = 0; lb < NUM_BLOCK_SURFS; lb++)
		{
			if (!light_chain[lb])
			{
				continue;
			}

			if (!light_surf[lb])
			{
				light_surf[lb] = CreateSurface(BLOCK_WIDTH, BLOCK_HEIGHT);
				block_changed[lb] = true;
			}
			if (block_changed[lb])
			{
				block_changed[lb] = false;
				buf = LockSurface(light_surf[lb]);
				memcpy(buf, light_block[lb], BLOCK_WIDTH * BLOCK_HEIGHT * 2);
				light_surf[lb]->Unlock(NULL);
			}

			RenderDevice->SetTexture(0, light_surf[lb]);

			for (cache = light_chain[lb]; cache; cache = cache->chain)
			{
				surf = cache->surf;
				tex = surf->texinfo;
				for (i = 0; i < surf->count; i++)
				{
					TVec texpt = surf->verts[i] - tex->texorg;
					s = DotProduct(texpt, tex->saxis);
					t = DotProduct(texpt, tex->taxis);
					lights = (s - surf->texturemins[0]) / 16 + cache->s + 0.5;
					lightt = (t - surf->texturemins[1]) / 16 + cache->t + 0.5;
					if (adjust_linear)
					{
						lights -= 0.5;
						lightt -= 0.5;
					}
					out[i] = D3DLVERTEX(D3DVECTOR(surf->verts[i].x,
						surf->verts[i].y, surf->verts[i].z), 0xffffffff, 0,
						lights / BLOCK_WIDTH, lightt / BLOCK_HEIGHT);
				}
				RenderDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, D3DFVF_LVERTEX, out, surf->count, 0);
			}
		}

		if (!tex_linear)
		{
			RenderDevice->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTFG_POINT);
			RenderDevice->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTFN_POINT);
		}
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, FALSE);
		RenderDevice->SetRenderState(D3DRENDERSTATE_SRCBLEND, D3DBLEND_SRCALPHA);
		RenderDevice->SetRenderState(D3DRENDERSTATE_DESTBLEND, D3DBLEND_INVSRCALPHA);
		RenderDevice->SetRenderState(D3DRENDERSTATE_ZWRITEENABLE, TRUE);		// back to normal Z buffering
	}
}

//==========================================================================
//
//	TDirect3DDrawer::DrawSkyPolygon
//
//==========================================================================

void TDirect3DDrawer::DrawSkyPolygon(TVec *cv, int count,
	int texture1, float offs1, int texture2, float offs2)
{
	D3DLVERTEX		out[256];
	int				i;

	viewData.dvMinZ = 0.99;
    RenderDevice->SetViewport(&viewData);

	if (r_use_fog)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_FOGENABLE, FALSE);
	}

	SetSkyTexture(texture1, false);
	for (i = 0; i < count; i++)
	{
		TVec v = cv[i] + vieworg;
		TVec texpt = cv[i] - r_texorg;
		out[i] = D3DLVERTEX(D3DVECTOR(v.x, v.y, v.z), 0xffffffff, 0,
			(DotProduct(texpt, r_saxis) - offs1) * tex_iw,
			DotProduct(texpt, r_taxis) * tex_ih);
		if (texAdjust)
		{
			out[i].tu -= 0.5 * tex_iw;
			out[i].tv -= 0.5 * tex_ih;
		}
	}
	RenderDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, D3DFVF_LVERTEX, out, count, 0);

	if (texture2)
	{
		SetSkyTexture(texture2, true);
		for (i = 0; i < count; i++)
		{
			TVec v = cv[i] + vieworg;
			TVec texpt = cv[i] - r_texorg;
			out[i] = D3DLVERTEX(D3DVECTOR(v.x, v.y, v.z), 0xffffffff, 0,
				(DotProduct(texpt, r_saxis) - offs2) * tex_iw,
				DotProduct(texpt, r_taxis) * tex_ih);
			if (texAdjust)
			{
				out[i].tu -= 0.5 * tex_iw;
				out[i].tv -= 0.5 * tex_ih;
			}
		}
		if (HavesAlphaTest)
		{
			RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, TRUE);
		}
		else
		{
			RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);
		}
		RenderDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, D3DFVF_LVERTEX, out, count, 0);
		if (HavesAlphaTest)
		{
			RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, FALSE);
		}
		else
		{
			RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, FALSE);
		}
	}

	if (r_use_fog)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_FOGENABLE, TRUE);
	}

	viewData.dvMinZ = 0;
    RenderDevice->SetViewport(&viewData);
}

//==========================================================================
//
//	TDirect3DDrawer::DrawMaskedPolygon
//
//==========================================================================

void TDirect3DDrawer::DrawMaskedPolygon(TVec *cv, int count,
	int texture, int translucency)
{
	D3DLVERTEX	out[256];
	int			i, r, g, b, alpha, w, h, size, l;
	surface_t	*surf = r_surface;

	if (texture & TEXF_FLAT)
	{
		SetFlat(texture);
	}
	else
	{
		SetTexture(texture);
	}

	R_BuildLightMap(surf, 0);
	w = (surf->extents[0] >> 4) + 1;
	h = (surf->extents[1] >> 4) + 1;
	size = w * h;
	r = 0;
	g = 0;
	b = 0;
	for (i = 0; i < size; i++)
	{
		r += 255 * 256 - blocklightsr[i];
		g += 255 * 256 - blocklightsg[i];
		b += 255 * 256 - blocklightsb[i];
	}
	r = r / (size * 256);
	g = g / (size * 256);
	b = b / (size * 256);
	alpha = (100 - translucency) * 255 / 100;
	l = (alpha << 24) | (r << 16) | (g << 8) | b;

	for (i = 0; i < count; i++)
	{
		TVec texpt = cv[i] - r_texorg;
		out[i] = D3DLVERTEX(D3DVECTOR(cv[i].x, cv[i].y, cv[i].z), l, 0,
			DotProduct(texpt, r_saxis) * tex_iw,
			DotProduct(texpt, r_taxis) * tex_ih);
		if (texAdjust)
		{
			out[i].tu -= 0.5 * tex_iw;
			out[i].tv -= 0.5 * tex_ih;
		}
	}

	if (HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, TRUE);
	}
	if (translucency || !HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);
	}

	RenderDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, D3DFVF_LVERTEX, out, count, 0);

	if (HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, FALSE);
	}
	if (translucency || !HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, FALSE);
	}
}

//==========================================================================
//
//	TDirect3DDrawer::DrawSpritePolygon
//
//==========================================================================

void TDirect3DDrawer::DrawSpritePolygon(TVec *cv, int lump,
	int translucency, int translation, dword light)
{													 
	D3DLVERTEX		out[4];
	int				i;

	SetSpriteLump(lump, translation);

	int l = (((100 - translucency) * 255 / 100) << 24) | (light & 0x00ffffff);
	for (i = 0; i < 4; i++)
	{
		TVec texpt = cv[i] - r_texorg;
		out[i] = D3DLVERTEX(D3DVECTOR(cv[i].x, cv[i].y, cv[i].z), l, 0,
			DotProduct(texpt, r_saxis) * tex_iw,
			DotProduct(texpt, r_taxis) * tex_ih);
		if (texAdjust)
		{
			out[i].tu -= 0.5 * tex_iw;
			out[i].tv -= 0.5 * tex_ih;
		}
	}

	if (translucency || !HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);
	}
	if (HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, TRUE);
	}

	RenderDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, D3DFVF_LVERTEX, out, 4, 0);

	if (translucency || !HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, FALSE);
	}
	if (HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, FALSE);
	}
}

//==========================================================================
//
//	TDirect3DDrawer::StartParticles
//
//==========================================================================

void TDirect3DDrawer::StartParticles(void)
{
	RenderDevice->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_DISABLE);
}

//==========================================================================
//
//	TDirect3DDrawer::DrawParticle
//
//==========================================================================

void TDirect3DDrawer::DrawParticle(particle_t *p)
{
	PART_VERTEX		out[4];
	out[0].origin = p->org;
	out[0].color = p->color;
	out[1].origin = p->org + viewright;
	out[1].color = p->color;
	out[2].origin = p->org + viewright - viewup;
	out[2].color = p->color;
	out[3].origin = p->org - viewup;
	out[3].color = p->color;
	RenderDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, PART_VERTEX_F, out, 4, 0);
}

//==========================================================================
//
//	TDirect3DDrawer::EndParticles
//
//==========================================================================

void TDirect3DDrawer::EndParticles(void)
{
	RenderDevice->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE);
}

