//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************
//**
//**	Functions to draw patches (by post) directly to screen.
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "d_local.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static picspanfunc_t	picspanfunc;
static byte				*picsource;

static int				ds_shade;

static byte				*picdata[MAX_PICS];
static int				picwidth[MAX_PICS];

// CODE --------------------------------------------------------------------

//==========================================================================
//
//	D_DrawPicSpan_8
//
//==========================================================================

void D_DrawPicSpan_8(int x, int y, fixed_t s, fixed_t t, fixed_t sstep, int count, byte *src)
{
	src += (t >> FRACBITS) * cachewidth;
    byte *dest = scrn + x + y * ScreenWidth;
	while (count--)
    {
		byte color = src[s >> FRACBITS];
		if (color)
		{
			*dest = color;
		}
		s += sstep;
		dest++;
	}
}

//==========================================================================
//
//	D_DrawPicSpanFuzz_8
//
//==========================================================================

void D_DrawPicSpanFuzz_8(int x, int y, fixed_t s, fixed_t t, fixed_t sstep, int count, byte *src)
{
	src += (t >> FRACBITS) * cachewidth;
    byte *dest = scrn + x + y * ScreenWidth;
	while (count--)
    {
		byte color = src[s >> FRACBITS];
		if (color)
		{
			*dest = ds_transluc[*dest + (color << 8)];
		}
		s += sstep;
		dest++;
	}
}

//==========================================================================
//
//	D_DrawPicSpanAltFuzz_8
//
//==========================================================================

void D_DrawPicSpanAltFuzz_8(int x, int y, fixed_t s, fixed_t t, fixed_t sstep, int count, byte *src)
{
	src += (t >> FRACBITS) * cachewidth;
    byte *dest = scrn + x + y * ScreenWidth;
	while (count--)
    {
		byte color = src[s >> FRACBITS];
		if (color)
		{
			*dest = ds_transluc[(*dest << 8) + color];
		}
		s += sstep;
		dest++;
	}
}

//==========================================================================
//
//	D_DrawPicSpanShadow_8
//
//==========================================================================

void D_DrawPicSpanShadow_8(int x, int y, fixed_t s, fixed_t t, fixed_t sstep, int count, byte *src)
{
	src += (t >> FRACBITS) * cachewidth;
    byte *dest = scrn + x + y * ScreenWidth;
	byte *cmap = colormaps + ((ds_shade >> 3) << 8);
	while (count--)
    {
		byte color = src[s >> FRACBITS];
		if (color)
		{
			*dest = cmap[*dest];
		}
		s += sstep;
		dest++;
	}
}

//==========================================================================
//
//	D_DrawFlatSpan_8
//
//==========================================================================

void D_DrawFlatSpan_8(int x, int y, fixed_t s, fixed_t t, fixed_t sstep, int count, byte *src)
{
	src += (t >> 10) & 0xfc0;
    byte *dest = scrn + x + y * ScreenWidth;
	while (count--)
    {
		byte color = src[(s >> FRACBITS) & 0x3f];
		*dest = color;
		s += sstep;
		dest++;
	}
}

//==========================================================================
//
//	D_ShadeRect_8
//
//  Fade all the screen buffer, so that the menu is more readable,
// especially now that we use the small hufont in the menus...
//
//==========================================================================

void D_ShadeRect_8(int xx, int yy, int ww, int hh, int darkening)
{
	int x1 = xx;
	int y1 = yy;
	int x2 = xx + ww;
	int y2 = yy + hh;
	if (x2 > ScreenWidth) x2 = ScreenWidth;
	if (y2 > ScreenHeight) y2 = ScreenHeight;

    byte *shades = colormaps + darkening * 256;
	for (int y = y1; y < y2; y++)
	{
		byte *dest = scrn + x1 + ScreenWidth * y;
		for (int x = x1; x < x2; x++)
		{
			*dest = shades[*dest];
			dest++;
		}
	}
}

//==========================================================================
//
//	D_DrawConsoleBackground_8
//
//==========================================================================

void D_DrawConsoleBackground_8(int h)
{
	static byte *consbgmap = NULL;
    if (!consbgmap)
	{
       	consbgmap = (byte*)W_CacheLumpName("CONSMAP", PU_STATIC);
	}

	int w = ScreenWidth >> 2;
	for (int y = 0; y < h; y++)
	{
		dword* buf = (dword*)(scrn + ScreenWidth * y);
		for (int x = 0; x < w; x++)
		{
   			dword quad = buf[x];
   			byte p1 = consbgmap[quad & 255];
   			byte p2 = consbgmap[(quad >> 8) & 255];
   			byte p3 = consbgmap[(quad >> 16) & 255];
   			byte p4 = consbgmap[quad >> 24];
   			buf[x] = (p4 << 24) | (p3 << 16) | (p2 << 8) | p1;
		 }
	}
}

//==========================================================================
//
//	D_DrawPicSpan_16
//
//==========================================================================

void D_DrawPicSpan_16(int x, int y, fixed_t s, fixed_t t, fixed_t sstep, int count, byte *src)
{
	src += (t >> FRACBITS) * cachewidth;
    word *dest = scrn16 + x + y * ScreenWidth;
	while (count--)
    {
		byte color = src[s >> FRACBITS];
		if (color)
		{
			*dest = pal8_to16[color];
		}
		s += sstep;
		dest++;
	}
}

//==========================================================================
//
//	D_DrawPicSpanFuzz_16
//
//==========================================================================

void D_DrawPicSpanFuzz_16(int x, int y, fixed_t s, fixed_t t, fixed_t sstep, int count, byte *src)
{
	src += (t >> FRACBITS) * cachewidth;
    word *dest = scrn16 + x + y * ScreenWidth;
	while (count--)
    {
		byte color = src[s >> FRACBITS];
		if (color)
		{
			word color16 = pal8_to16[color];
			byte r1 = GetColR(*dest);
			byte g1 = GetColG(*dest);
			byte b1 = GetColB(*dest);
			byte r2 = GetColR(color16);
			byte g2 = GetColG(color16);
			byte b2 = GetColB(color16);
			byte r = (r1 * ds_transluc16 + r2 * (100 - ds_transluc16)) / 100;
			byte g = (g1 * ds_transluc16 + g2 * (100 - ds_transluc16)) / 100;
			byte b = (b1 * ds_transluc16 + b2 * (100 - ds_transluc16)) / 100;
			*dest = MakeCol(r, g, b);
		}
		s += sstep;
		dest++;
	}
}

//==========================================================================
//
//	D_DrawPicSpanShadow_16
//
//==========================================================================

void D_DrawPicSpanShadow_16(int x, int y, fixed_t s, fixed_t t, fixed_t sstep, int count, byte *src)
{
	src += (t >> FRACBITS) * cachewidth;
    word *dest = scrn16 + x + y * ScreenWidth;
	int ishade = 255 - ds_shade;
	while (count--)
    {
		byte color = src[s >> FRACBITS];
		if (color)
		{
			byte r = GetColR(*dest);
			byte g = GetColG(*dest);
			byte b = GetColB(*dest);
			*dest = MakeCol(r * ishade / 256, g * ishade / 256, b * ishade / 256);
		}
		s += sstep;
		dest++;
	}
}

//==========================================================================
//
//	D_DrawFlatSpan_16
//
//==========================================================================

void D_DrawFlatSpan_16(int x, int y, fixed_t s, fixed_t t, fixed_t sstep, int count, byte *src)
{
	src += (t >> 10) & 0xfc0;
    word *dest = scrn16 + x + y * ScreenWidth;
	while (count--)
    {
		byte color = src[(s >> FRACBITS) & 0x3f];
		*dest = pal8_to16[color];
		s += sstep;
		dest++;
	}
}

//==========================================================================
//
//	D_ShadeRect_16
//
//  Fade all the screen buffer, so that the menu is more readable,
// especially now that we use the small hufont in the menus...
//
//==========================================================================

void D_ShadeRect_16(int xx, int yy, int ww, int hh, int darkening)
{
	int x1 = xx;
	int y1 = yy;
	int x2 = xx + ww;
	int y2 = yy + hh;
	if (x2 > ScreenWidth) x2 = ScreenWidth;
	if (y2 > ScreenHeight) y2 = ScreenHeight;

	darkening = 32 - darkening;
	for (int y = y1; y < y2; y++)
	{
		word *dest = scrn16 + x1 + ScreenWidth * y;
		for (int x = x1; x < x2; x++)
		{
			byte r = GetColR(*dest);
			byte g = GetColG(*dest);
			byte b = GetColB(*dest);
			*dest = MakeCol(r * darkening / 32, g * darkening / 32, b * darkening / 32);
			dest++;
		}
	}
}

//==========================================================================
//
//	D_DrawConsoleBackground_16
//
//==========================================================================

void D_DrawConsoleBackground_16(int h)
{
	for (int y = 0; y < h; y++)
	{
		word *dest = scrn16 + ScreenWidth * y;
		for (int x = 0; x < ScreenWidth; x++)
		{
			byte r = GetColR(*dest);
			byte g = GetColG(*dest);
			byte b = GetColB(*dest);
			byte col = (r * 5 + g * 8 + b * 3) / 16;
			*dest = MakeCol(0, 0, col);
			dest++;
		}
	}
}

//==========================================================================
//
//	D_DrawPicSpan_32
//
//==========================================================================

void D_DrawPicSpan_32(int x, int y, fixed_t s, fixed_t t, fixed_t sstep, int count, byte *src)
{
	src += (t >> FRACBITS) * cachewidth;
    dword *dest = (dword*)scrn + x + y * ScreenWidth;
	while (count--)
    {
		byte color = src[s >> FRACBITS];
		if (color)
		{
			*dest = pal2rgb[color];
		}
		s += sstep;
		dest++;
	}
}

//==========================================================================
//
//	D_DrawPicSpanFuzz_32
//
//==========================================================================

void D_DrawPicSpanFuzz_32(int x, int y, fixed_t s, fixed_t t, fixed_t sstep, int count, byte *src)
{
	src += (t >> FRACBITS) * cachewidth;
    dword *dest = (dword*)scrn + x + y * ScreenWidth;
	while (count--)
    {
		byte color = src[s >> FRACBITS];
		if (color)
		{
			dword color32 = pal2rgb[color];
			byte r1 = GetCol32R(*dest);
			byte g1 = GetCol32G(*dest);
			byte b1 = GetCol32B(*dest);
			byte r2 = GetCol32R(color32);
			byte g2 = GetCol32G(color32);
			byte b2 = GetCol32B(color32);
			byte r = (r1 * ds_transluc16 + r2 * (100 - ds_transluc16)) / 100;
			byte g = (g1 * ds_transluc16 + g2 * (100 - ds_transluc16)) / 100;
			byte b = (b1 * ds_transluc16 + b2 * (100 - ds_transluc16)) / 100;
			*dest = MakeCol32(r, g, b);
		}
		s += sstep;
		dest++;
	}
}

//==========================================================================
//
//	D_DrawPicSpanShadow_32
//
//==========================================================================

void D_DrawPicSpanShadow_32(int x, int y, fixed_t s, fixed_t t, fixed_t sstep, int count, byte *src)
{
	src += (t >> FRACBITS) * cachewidth;
    dword *dest = (dword*)scrn + x + y * ScreenWidth;
	int ishade = 255 - ds_shade;
	while (count--)
    {
		byte color = src[s >> FRACBITS];
		if (color)
		{
			byte r = GetCol32R(*dest);
			byte g = GetCol32G(*dest);
			byte b = GetCol32B(*dest);
			*dest = MakeCol32(r * ishade / 256, g * ishade / 256, b * ishade / 256);
		}
		s += sstep;
		dest++;
	}
}

//==========================================================================
//
//	D_DrawFlatSpan_32
//
//==========================================================================

void D_DrawFlatSpan_32(int x, int y, fixed_t s, fixed_t t, fixed_t sstep, int count, byte *src)
{
	src += (t >> 10) & 0xfc0;
    dword *dest = (dword*)scrn + x + y * ScreenWidth;
	while (count--)
    {
		byte color = src[(s >> FRACBITS) & 0x3f];
		*dest = pal2rgb[color];
		s += sstep;
		dest++;
	}
}

//==========================================================================
//
//	D_ShadeRect_32
//
//  Fade all the screen buffer, so that the menu is more readable,
// especially now that we use the small hufont in the menus...
//
//==========================================================================

void D_ShadeRect_32(int xx, int yy, int ww, int hh, int darkening)
{
	int x1 = xx;
	int y1 = yy;
	int x2 = xx + ww;
	int y2 = yy + hh;
	if (x2 > ScreenWidth) x2 = ScreenWidth;
	if (y2 > ScreenHeight) y2 = ScreenHeight;

	darkening = 32 - darkening;
	for (int y = y1; y < y2; y++)
	{
		dword *dest = (dword*)scrn + x1 + ScreenWidth * y;
		for (int x = x1; x < x2; x++)
		{
			byte r = GetCol32R(*dest);
			byte g = GetCol32G(*dest);
			byte b = GetCol32B(*dest);
			*dest = MakeCol32(r * darkening / 32, g * darkening / 32, b * darkening / 32);
			dest++;
		}
	}
}

//==========================================================================
//
//	D_DrawConsoleBackground_32
//
//==========================================================================

void D_DrawConsoleBackground_32(int h)
{
	for (int y = 0; y < h; y++)
	{
		dword *dest = (dword*)scrn + ScreenWidth * y;
		for (int x = 0; x < ScreenWidth; x++)
		{
			byte r = GetCol32R(*dest);
			byte g = GetCol32G(*dest);
			byte b = GetCol32B(*dest);
			byte col = (r * 5 + g * 8 + b * 3) / 16;
			*dest = MakeCol32(0, 0, col);
			dest++;
		}
	}
}

//==========================================================================
//
//	SetPatch
//
//==========================================================================

static void GeneratePicFromPatch(int handle)
{
	patch_t *patch = (patch_t*)W_CacheLumpName(pic_list[handle].name, PU_TEMP);
	int w = LittleShort(patch->width);
	int h = LittleShort(patch->height);
	byte *block = (byte*)Z_Calloc(w * h, PU_CACHE, (void**)&picdata[handle]);
	picdata[handle] = block;
	picwidth[handle] = w;

	for (int x = 0; x < w; x++)
	{
    	column_t *column = (column_t *)((byte *)patch + LittleLong(patch->columnofs[x]));

		// step through the posts in a column
	    while (column->topdelta != 0xff)
	    {
		    byte* source = (byte *)column + 3;
		    byte* dest = block + x + column->topdelta * w;
			int count = column->length;

	    	while (count--)
	    	{
				*dest = *source ? *source : black_color;
				source++;
				dest += w;
	    	}
			column = (column_t *)((byte *)column + column->length + 4);
	    }
	}
}

//==========================================================================
//
//	GeneratePicFromRaw
//
//==========================================================================

void GeneratePicFromRaw(int handle)
{
	picdata[handle] = (byte*)Z_Malloc(320 * 200, PU_CACHE, (void**)&picdata[handle]);
	W_ReadLump(W_GetNumForName(pic_list[handle].name), picdata[handle]);
	for (int i = 0; i < 320 * 200; i++)
	{
		if (!picdata[handle][i])
		{
			picdata[handle][i] = black_color;
		}
	}
	picwidth[handle] = 320;
}

//==========================================================================
//
//	SetPic
//
//==========================================================================

void SetPic(int handle)
{
	if (!picdata[handle])
	{
		switch (pic_list[handle].type)
 		{
	 	 case PIC_PATCH:
			GeneratePicFromPatch(handle);
			break;

		 case PIC_RAW:
			GeneratePicFromRaw(handle);
			break;
		}
	}
	picsource = picdata[handle];
	cachewidth = picwidth[handle];
}

//==========================================================================
//
//	D_DrawPic
//
//==========================================================================

void D_DrawPic(float x1, float x2, float y1, float y2,
	float s1, float s2, float t1, float t2)
{
	if (x1 < 0)
	{
		s1 = s1 + x1 / (x1 - x2) * (s2 - s1);
		x1 = 0;
	}
	if (x2 > ScreenWidth)
	{
		s2 = s2 + (x2 - ScreenWidth) / (x1 - x2) * (s2 - s1);
		x2 = ScreenWidth;
	}
	if (y1 < 0)
	{
		t1 = t1 + y1 / (y1 - y2) * (t2 - t1);
		y1 = 0;
	}
	if (y2 > ScreenHeight)
	{
		t2 = t2 + (y2 - ScreenHeight) / (y1 - y2) * (t2 - t1);
		y2 = ScreenHeight;
	}

	int top = (int)y1;
	int bot = (int)y2;
	int lines = bot - top;
	if (lines <= 0)
	{
		return;
	}

	int left = (int)x1;
	int right = (int)x2;
	int count = right - left;
	if (count <= 0)
	{
		return;
	}

	fixed_t sstep = FX((s2 - s1) / (x2 - x1));
	fixed_t tstep = FX((t2 - t1) / (y2 - y1));

	fixed_t s = FX(s1);
	fixed_t t = FX(t1);

	int iy = top;

	do
	{
		picspanfunc(left, iy, s, t, sstep, count, picsource);
		iy++;
		t += tstep;
	}
	while (--lines);
}

//==========================================================================
//
//	TSoftwareDrawer::DrawPic
//
//==========================================================================

void TSoftwareDrawer::DrawPic(float x1, float x2, float y1, float y2,
	float s1, float s2, float t1, float t2, int handle, int trans)
{
	SetPic(handle);
	if (ScreenBPP == 8)
	{
		if (trans < 5)
		{
			picspanfunc = D_DrawPicSpan_8;
		}
		else if (trans < 45)
		{
			ds_transluc = tinttables[(trans - 5) / 10];
			picspanfunc = D_DrawPicSpanFuzz_8;
		}
		else if (trans < 95)
		{
			ds_transluc = tinttables[(94 - trans) / 10];
			picspanfunc = D_DrawPicSpanAltFuzz_8;
		}
		else
		{
			return;
		}
	}
	else if (PixelBytes == 2)
	{
		if (trans <= 0)
		{
			picspanfunc = D_DrawPicSpan_16;
		}
		else if (trans < 100)
		{
			ds_transluc16 = trans;
			picspanfunc = D_DrawPicSpanFuzz_16;
		}
		else
		{
			return;
		}
	}
	else
	{
		if (trans <= 0)
		{
			picspanfunc = D_DrawPicSpan_32;
		}
		else if (trans < 100)
		{
			ds_transluc16 = trans;
			picspanfunc = D_DrawPicSpanFuzz_32;
		}
		else
		{
			return;
		}
	}
	D_DrawPic(x1, x2, y1, y2, s1, s2, t1, t2);
}

//==========================================================================
//
//	TSoftwareDrawer::DrawPicShadow
//
//==========================================================================

void TSoftwareDrawer::DrawPicShadow(float x1, float x2, float y1, float y2,
	float s1, float s2, float t1, float t2, int handle, int shade)
{
	SetPic(handle);
	ds_shade = shade;
	if (ScreenBPP == 8)
	{
		picspanfunc = D_DrawPicSpanShadow_8;
	}
	else if (PixelBytes == 2)
	{
		picspanfunc = D_DrawPicSpanShadow_16;
	}
	else
	{
		picspanfunc = D_DrawPicSpanShadow_32;
	}
	D_DrawPic(x1, x2, y1, y2, s1, s2, t1, t2);
}

//==========================================================================
//
//  TSoftwareDrawer::FillRectWithFlat
//
// 	Fills rectangle with flat.
//
//==========================================================================

void TSoftwareDrawer::FillRectWithFlat(int DestX, int DestY, int width, int height, const char* fname)
{
	SetFlat(R_FlatNumForName(fname));
	picsource = (byte*)cacheblock;
	if (ScreenBPP == 8)
	{
		picspanfunc = D_DrawFlatSpan_8;
	}
	else if (PixelBytes == 2)
	{
		picspanfunc = D_DrawFlatSpan_16;
	}
	else
	{
		picspanfunc = D_DrawFlatSpan_32;
	}
	D_DrawPic(DestX, DestX + width, DestY, DestY + height, 0, width, 0, height);
}

//==========================================================================
//
//	TSoftwareDrawer::DarkenScreen
//
//  Fade all the screen buffer, so that the menu is more readable,
// especially now that we use the small hufont in the menus...
//
//==========================================================================

void TSoftwareDrawer::ShadeRect(int x, int y, int w, int h, int darkening)
{
	if (ScreenBPP == 8)
		D_ShadeRect_8(x, y, w, h, darkening);
	else if (PixelBytes == 2)
		D_ShadeRect_16(x, y, w, h, darkening);
	else
		D_ShadeRect_32(x, y, w, h, darkening);
}

//==========================================================================
//
//	TSoftwareDrawer::DrawConsoleBackground
//
//==========================================================================

void TSoftwareDrawer::DrawConsoleBackground(int h)
{
	if (ScreenBPP == 8)
		D_DrawConsoleBackground_8(h);
	else if (PixelBytes == 2)
		D_DrawConsoleBackground_16(h);
	else
		D_DrawConsoleBackground_32(h);
}

//==========================================================================
//
//	TSoftwareDrawer::DrawSpriteLump
//
//==========================================================================

void TSoftwareDrawer::DrawSpriteLump(int x, int y, int lump, int translation, boolean flip)
{
	SetSpriteLump(lump, 0, translation);
	picsource = (byte*)cacheblock;
    x -= (int)spriteoffset[lump];
   	y -= (int)spritetopoffset[lump];
	picspanfunc = ScreenBPP == 8 ? D_DrawPicSpan_8 : D_DrawPicSpan_16;
	float w = spritewidth[lump];
	float h = spriteheight[lump];

    if (flip)
		D_DrawPic(fScaleX * x, fScaleX * (x + w), fScaleY * y,
			fScaleY * (y + h), w - 0.0001, 0, 0, h);
    else
		D_DrawPic(fScaleX * x, fScaleX * (x + w), fScaleY * y,
			fScaleY * (y + h), 0, w, 0, h);
}


