//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "gl_local.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

//==========================================================================
//
//	TOpenGLDrawer::FlushCaches
//
//==========================================================================

void TOpenGLDrawer::FlushCaches(bool free_blocks)
{
	int				i;
	surfcache_t		*blines;
	surfcache_t		*block;

	if (free_blocks)
	{
		for (i = 0; i < NUM_BLOCK_SURFS; i++)
		{
			for (blines = cacheblocks[i]; blines; blines = blines->bnext)
			{
				for (block = blines; block; block = block->lnext)
				{
					if (block->owner)
						*block->owner = NULL;
				}
			}
		}
	}

	memset(blockbuf, 0, sizeof(blockbuf));
	freeblocks = NULL;
	for (i = 0; i < NUM_CACHE_BLOCKS; i++)
	{
		blockbuf[i].chain = freeblocks;
		freeblocks = &blockbuf[i];
	}
	for (i = 0; i < NUM_BLOCK_SURFS; i++)
	{
		cacheblocks[i] = freeblocks;
		freeblocks = freeblocks->chain;
		cacheblocks[i]->width = BLOCK_WIDTH;
		cacheblocks[i]->height = BLOCK_HEIGHT;
		cacheblocks[i]->blocknum = i;
	}
}

//==========================================================================
//
//	TOpenGLDrawer::AllocBlock
//
//==========================================================================

surfcache_t	*TOpenGLDrawer::AllocBlock(int width, int height)
{
	int				i;
	surfcache_t		*blines;
	surfcache_t		*block;
	surfcache_t		*other;

	for (i = 0; i < NUM_BLOCK_SURFS; i++)
	{
		for (blines = cacheblocks[i]; blines; blines = blines->bnext)
		{
			if (blines->height != height)
			{
				continue;
			}
			for (block = blines; block; block = block->lnext)
			{
				if (block->owner)
					continue;
				if (block->width < width)
					continue;
				if (block->width > width)
				{
					if (!freeblocks)
						Sys_Error("No more free blocks");
					other = freeblocks;
					freeblocks = other->chain;
					other->s = block->s + width;
					other->t = block->t;
					other->width = block->width - width;
					other->height = block->height;
					other->lnext = block->lnext;
					if (other->lnext)
						other->lnext->lprev = other;
					block->lnext = other;
					other->lprev = block;
					block->width = width;
					other->owner = NULL;
					other->blocknum = i;
				}
				return block;
			}
		}
	}

	for (i = 0; i < NUM_BLOCK_SURFS; i++)
	{
		for (blines = cacheblocks[i]; blines; blines = blines->bnext)
		{
			if (blines->height < height)
			{
				continue;
			}
			if (blines->lnext)
			{
				continue;
			}

			block = blines;
			if (block->height > height)
			{
				if (!freeblocks)
					Sys_Error("No more free blocks");
				other = freeblocks;
				freeblocks = other->chain;
				other->s = 0;
				other->t = block->t + height;
				other->width = block->width;
				other->height = block->height - height;
				other->lnext = NULL;
				other->lprev = NULL;
				other->bnext = block->bnext;
				if (other->bnext)
					other->bnext->bprev = other;
				block->bnext = other;
				other->bprev = block;
				block->height = height;
				other->owner = NULL;
				other->blocknum = i;
			}

			if (!freeblocks)
				Sys_Error("No more free blocks");
			other = freeblocks;
			freeblocks = other->chain;
			other->s = block->s + width;
			other->t = block->t;
			other->width = block->width - width;
			other->height = block->height;
			other->lnext = NULL;
			block->lnext = other;
			other->lprev = block;
			block->width = width;
			other->owner = NULL;
			other->blocknum = i;

			return block;
		}
	}

	Sys_Error("overflow");
	return NULL;
}

//==========================================================================
//
//	TOpenGLDrawer::FreeSurfCache
//
//==========================================================================

void TOpenGLDrawer::FreeSurfCache(surfcache_t *block)
{
	surfcache_t		*other;

	*block->owner = NULL;
	block->owner = NULL;
	if (block->lnext && !block->lnext->owner)
	{
		other = block->lnext;
		block->width += other->width;
		block->lnext = other->lnext;
		if (block->lnext)
			block->lnext->lprev = block;
		other->chain = freeblocks;
		freeblocks = other;
	}
	if (block->lprev && !block->lprev->owner)
	{
		other = block;
		block = block->lprev;
		block->width += other->width;
		block->lnext = other->lnext;
		if (block->lnext)
			block->lnext->lprev = block;
		other->chain = freeblocks;
		freeblocks = other;
	}

	if (block->lprev || block->lnext)
	{
		return;
	}

	if (block->bnext && !block->bnext->lnext)
	{
		other = block->bnext;
		block->height += other->height;
		block->bnext = other->bnext;
		if (block->bnext)
			block->bnext->bprev = block;
		other->chain = freeblocks;
		freeblocks = other;
	}
	if (block->bprev && !block->bprev->lnext)
	{
		other = block;
		block = block->bprev;
		block->height += other->height;
		block->bnext = other->bnext;
		if (block->bnext)
			block->bnext->bprev = block;
		other->chain = freeblocks;
		freeblocks = other;
	}
}

//==========================================================================
//
//	TOpenGLDrawer::CacheSurface
//
//==========================================================================

void TOpenGLDrawer::CacheSurface(surface_t *surface)
{
	surfcache_t     *cache;
	int				smax, tmax;
	int				i, j, bnum;

	//
	// see if the cache holds apropriate data
	//
	cache = surface->cachespots[0];

	if (cache && !cache->dlight && surface->dlightframe != r_dlightframecount
			&& cache->lightlevel == surface->lightlevel)
	{
		bnum = cache->blocknum;
		cache->chain = light_chain[bnum];
		light_chain[bnum] = cache;
		return;
	}

	//
	// determine shape of surface
	//
	smax = (surface->extents[0] >> 4) + 1;
	tmax = (surface->extents[1] >> 4) + 1;
	
	//
	// allocate memory if needed
	//
	if (!cache)     // if a texture just animated, don't reallocate it
	{
		cache = AllocBlock(smax, tmax);
		surface->cachespots[0] = cache;
		cache->owner = &surface->cachespots[0];
		cache->surf = surface;
	}
	
	if (surface->dlightframe == r_dlightframecount)
		cache->dlight = 1;
	else
		cache->dlight = 0;
	cache->lightlevel = surface->lightlevel;

	// calculate the lightings
	R_BuildLightMap(surface, 0);
	bnum = cache->blocknum;
	block_changed[bnum] = true;

	for (j = 0; j < tmax; j++)
	{
		for (i = 0; i < smax; i++)
		{
			light_block[bnum][(j + cache->t) * BLOCK_WIDTH + i + cache->s].r = blocklightsr[j * smax + i] >> 8;
			light_block[bnum][(j + cache->t) * BLOCK_WIDTH + i + cache->s].g = blocklightsg[j * smax + i] >> 8;
			light_block[bnum][(j + cache->t) * BLOCK_WIDTH + i + cache->s].b = blocklightsb[j * smax + i] >> 8;
			light_block[bnum][(j + cache->t) * BLOCK_WIDTH + i + cache->s].a = 255;
		}
	}
	cache->chain = light_chain[bnum];
	light_chain[bnum] = cache;
}

//==========================================================================
//
//	glVertex
//
//==========================================================================

inline void	glVertex(const TVec &v)
{
	glVertex3f(v.x, v.y, v.z);
}

//==========================================================================
//
//	TOpenGLDrawer::DrawPolygon
//
//==========================================================================

void TOpenGLDrawer::DrawPolygon(TVec *cv, int count, int texture, int)
{
	int			i;
	bool		lightmaped;
	surface_t	*surf = r_surface;

	if (texture & TEXF_FLAT)
	{
		SetFlat(texture);
	}
	else
	{
		SetTexture(texture);
	}

	lightmaped = surf->lightmap != NULL ||
		surf->dlightframe == r_dlightframecount;

	if (lightmaped)
	{
		glColor4f(1, 1, 1, 1);
	}
	else
	{
		glColor4ub(surf->lightlevel, surf->lightlevel, surf->lightlevel, 255);
	}

	glBegin(GL_POLYGON);
	for (i = 0; i < count; i++)
	{
		TVec texpt = cv[i] - r_texorg;
		glTexCoord2f(DotProduct(texpt, r_saxis) * tex_iw,
			DotProduct(texpt, r_taxis) * tex_ih);
		glVertex(cv[i]);
	}
	glEnd();

	if (lightmaped)
	{
		CacheSurface(surf);
	}
}

//==========================================================================
//
//	TOpenGLDrawer::WorldDrawing
//
//==========================================================================

void TOpenGLDrawer::WorldDrawing(void)
{
	int			lb, i;
	surfcache_t	*cache;
	float		s, t;
	surface_t	*surf;
	texinfo_t	*tex;

	glDepthMask(0);		// don't bother writing Z
	glBlendFunc(GL_ZERO, GL_ONE_MINUS_SRC_COLOR);
	glEnable(GL_BLEND);
	glColor4f(1, 1, 1, 1);

	for (lb = 0; lb < NUM_BLOCK_SURFS; lb++)
	{
		if (!light_chain[lb])
		{
			continue;
		}

		glBindTexture(GL_TEXTURE_2D, lmap_id[lb]);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		if (block_changed[lb])
		{
			block_changed[lb] = false;
			glTexImage2D(GL_TEXTURE_2D, 0, 4, BLOCK_WIDTH, BLOCK_HEIGHT,
				0, GL_RGBA, GL_UNSIGNED_BYTE, light_block[lb]);
		}

		for (cache = light_chain[lb]; cache; cache = cache->chain)
		{
			surf = cache->surf;
			tex = surf->texinfo;

			glBegin(GL_POLYGON);
			for (i = 0; i < surf->count; i++)
			{
				TVec texpt = surf->verts[i] - tex->texorg;
				s =	(DotProduct(texpt, tex->saxis) - surf->texturemins[0] +
					cache->s * 16 + 8) / (BLOCK_WIDTH * 16);
				t =	(DotProduct(texpt, tex->taxis) - surf->texturemins[1] +
					cache->t * 16 + 8) / (BLOCK_HEIGHT * 16);
				glTexCoord2f(s, t);
				glVertex(surf->verts[i]);
			}
			glEnd();
		}
	}

	glDisable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDepthMask(1);		// back to normal Z buffering
}

//==========================================================================
//
//	TOpenGLDrawer::DrawSkyPolygon
//
//==========================================================================

void TOpenGLDrawer::DrawSkyPolygon(TVec *cv, int count,
	int texture1, float offs1, int texture2, float offs2)
{
	int		i;

	glDepthRange(0.99, 1.0);

	//	Sky polys are alredy translated
	glPushMatrix();
	glTranslatef(vieworg.x, vieworg.y, vieworg.z);

	if (r_use_fog)
	{
		glDisable(GL_FOG);
	}

	SetSkyTexture(texture1, false);
	glBegin(GL_POLYGON);
	glColor4f(1, 1, 1, 1);
	for (i = 0; i < count; i++)
	{
		TVec texpt = cv[i] - r_texorg;
		glTexCoord2f((DotProduct(texpt, r_saxis) - offs1) * tex_iw,
			DotProduct(texpt, r_taxis) * tex_ih);
		glVertex(cv[i]);
	}
	glEnd();

	if (texture2)
	{
		SetSkyTexture(texture2, true);
		glEnable(GL_ALPHA_TEST);
		glBegin(GL_POLYGON);
		glColor4f(1, 1, 1, 1);
		for (i = 0; i < count; i++)
		{
			TVec texpt = cv[i] - r_texorg;
			glTexCoord2f((DotProduct(texpt, r_saxis) - offs2) * tex_iw,
				DotProduct(texpt, r_taxis) * tex_ih);
			glVertex(cv[i]);
		}
		glEnd();
		glDisable(GL_ALPHA_TEST);
	}

	if (r_use_fog)
	{
		glEnable(GL_FOG);
	}

	glPopMatrix();
	glDepthRange(0.0, 1.0);
}

//==========================================================================
//
//	TOpenGLDrawer::DrawMaskedPolygon
//
//==========================================================================

void TOpenGLDrawer::DrawMaskedPolygon(TVec *cv, int count,
	int texture, int translucency)
{
	int			i, r, g, b, w, h, size;
	double		iscale;
	surface_t	*surf = r_surface;

	if (texture & TEXF_FLAT)
	{
		SetFlat(texture);
	}
	else
	{
		SetTexture(texture);
	}
	glEnable(GL_ALPHA_TEST);
	if (translucency)
	{
		glEnable(GL_BLEND);
	}

	R_BuildLightMap(surf, 0);
	w = (surf->extents[0] >> 4) + 1;
	h = (surf->extents[1] >> 4) + 1;
	size = w * h;
	r = 0;
	g = 0;
	b = 0;
	for (i = 0; i < size; i++)
	{
		r += 255 * 256 - blocklightsr[i];
		g += 255 * 256 - blocklightsg[i];
		b += 255 * 256 - blocklightsb[i];
	}
	float alpha = (100.0 - translucency) / 100.0;
	iscale = 1.0 / (size * 255 * 256);
	glColor4f(r * iscale, g * iscale, b * iscale, alpha);

	glBegin(GL_POLYGON);
	for (i = 0; i < count; i++)
	{
		TVec texpt = cv[i] - r_texorg;
		glTexCoord2f(DotProduct(texpt, r_saxis) * tex_iw,
			DotProduct(texpt, r_taxis) * tex_ih);
		glVertex(cv[i]);
	}
	glEnd();

	if (translucency)
	{
		glDisable(GL_BLEND);
	}
	glDisable(GL_ALPHA_TEST);
}

//==========================================================================
//
//	TOpenGLDrawer::DrawSpritePolygon
//
//==========================================================================

void TOpenGLDrawer::DrawSpritePolygon(TVec *cv, int lump,
	int translucency, int translation, dword light)
{
	TVec	texpt;

	SetSpriteLump(lump, translation);

	if (translucency)
	{
		glEnable(GL_BLEND);
	}
	glEnable(GL_ALPHA_TEST);

	int alpha = 255 * (100 - translucency) / 100;
	glColor4ub(light >> 16, light >> 8, light, alpha);

	glBegin(GL_QUADS);

	texpt = cv[0] - r_texorg;
	glTexCoord2f(DotProduct(texpt, r_saxis) * tex_iw,
		DotProduct(texpt, r_taxis) * tex_ih);
	glVertex(cv[0]);

	texpt = cv[1] - r_texorg;
	glTexCoord2f(DotProduct(texpt, r_saxis) * tex_iw,
		DotProduct(texpt, r_taxis) * tex_ih);
	glVertex(cv[1]);

	texpt = cv[2] - r_texorg;
	glTexCoord2f(DotProduct(texpt, r_saxis) * tex_iw,
		DotProduct(texpt, r_taxis) * tex_ih);
	glVertex(cv[2]);

	texpt = cv[3] - r_texorg;
	glTexCoord2f(DotProduct(texpt, r_saxis) * tex_iw,
		DotProduct(texpt, r_taxis) * tex_ih);
	glVertex(cv[3]);

	glEnd();

	if (translucency)
	{
		glDisable(GL_BLEND);
	}
	glDisable(GL_ALPHA_TEST);
}

//==========================================================================
//
//	TOpenGLDrawer::StartParticles
//
//==========================================================================

void TOpenGLDrawer::StartParticles(void)
{
	glDisable(GL_TEXTURE_2D);
	glBegin(GL_QUADS);
}

//==========================================================================
//
//	TOpenGLDrawer::DrawParticle
//
//==========================================================================

void TOpenGLDrawer::DrawParticle(particle_t *p)
{
	glColor4ub((p->color >> 16) & 0xff, (p->color >> 8) & 0xff,
		p->color & 0xff, p->color >> 24);
	glVertex(p->org);
	glVertex(p->org + viewright);
	glVertex(p->org + viewright - viewup);
	glVertex(p->org - viewup);
}

//==========================================================================
//
//	TOpenGLDrawer::EndParticles
//
//==========================================================================

void TOpenGLDrawer::EndParticles(void)
{
	glEnd();
	glEnable(GL_TEXTURE_2D);
}

