//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "gl_local.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

//==========================================================================
//
//	TOpenGLDrawer::InitData
//
//==========================================================================

void TOpenGLDrawer::InitData(void)
{
	byte *pal = (byte*)W_CacheLumpName("PLAYPAL", PU_CACHE);
	for (int i = 0; i < 256; i++)
	{
		pal8_to24[i].r = *pal++;
		pal8_to24[i].g = *pal++;
		pal8_to24[i].b = *pal++;
		pal8_to24[i].a = 255;
	}
}

//==========================================================================
//
//	TOpenGLDrawer::InitTextures
//
//==========================================================================

void TOpenGLDrawer::InitTextures(void)
{
	//	Textures
	texture_id = (GLuint*)Z_Calloc(numtextures * 4);
	texture_sent = (bool*)Z_Calloc(numtextures);
	texture_iw = (float*)Z_Calloc(numtextures * 4);
	texture_ih = (float*)Z_Calloc(numtextures * 4);
	// 	Flats
	flat_id = (GLuint*)Z_Calloc(numflats * 4);
	flat_sent = (bool*)Z_Calloc(numflats);
	// 	Sprite lumps
    sprite_id = (GLuint*)Z_Calloc(numspritelumps * 4);
    sprite_sent = (bool*)Z_Calloc(numspritelumps);
    spriteiw = (float*)Z_Calloc(numspritelumps * 4);
    spriteih = (float*)Z_Calloc(numspritelumps * 4);
}

//==========================================================================
//
//	TOpenGLDrawer::GenerateTextures
//
//==========================================================================

void TOpenGLDrawer::GenerateTextures(void)
{
	glGenTextures(numtextures, texture_id);
	glGenTextures(numflats, flat_id);
	glGenTextures(numspritelumps, sprite_id);
	glGenTextures(MAX_TRANSLATED_SPRITES, trspr_id);
	glGenTextures(MAX_PICS, pic_id);
	glGenTextures(NUM_BLOCK_SURFS, lmap_id);
	memset(texture_sent, 0, numtextures);
	memset(flat_sent, 0, numflats);
	memset(sprite_sent, 0, numspritelumps);
	memset(trspr_id, 0, MAX_TRANSLATED_SPRITES);
	memset(pic_sent, 0, MAX_PICS);
	texturesGenerated = true;
}

//==========================================================================
//
//	TOpenGLDrawer::DeleteTextures
//
//==========================================================================

void TOpenGLDrawer::DeleteTextures(void)
{
	if (texturesGenerated)
	{
		glDeleteTextures(numtextures, texture_id);
		glDeleteTextures(numflats, flat_id);
		glDeleteTextures(numspritelumps, sprite_id);
		glDeleteTextures(MAX_TRANSLATED_SPRITES, trspr_id);
		glDeleteTextures(MAX_PICS, pic_id);
		glDeleteTextures(NUM_BLOCK_SURFS, lmap_id);
		texturesGenerated = false;
	}
}

//==========================================================================
//
//	TOpenGLDrawer::ToPowerOf2
//
//==========================================================================

int TOpenGLDrawer::ToPowerOf2(int val)
{
	int answer = 1;
	while (answer < val)
		answer <<= 1;
	return answer;
}

//==========================================================================
//
//	TOpenGLDrawer::DrawColumnInCache
//
// 	Clip and draw a column from a patch into a flat buffer.
//
//		column - column to draw
//		cache - buffer
//		originx, originy - position of column in the buffer
//		cachewidth, cacheheight - size of the cache
//
//==========================================================================

void TOpenGLDrawer::DrawColumnInCache(column_t* column, rgba_t* cache,
	int originx, int originy, int cachewidth, int cacheheight)
{
	int		count;
	int		position;
	byte*	source;
	rgba_t*	dest;

	// step through the posts in a column
	while (column->topdelta != 0xff)
	{
		source = (byte *)column + 3;
		count = column->length;
		position = originy + column->topdelta;

		//	Clip position
		if (position < 0)
		{
			count += position;
			source -= position;
			position = 0;
		}
		if (position + count > cacheheight)
		{
			count = cacheheight - position;
		}
		dest = cache + originx + position * cachewidth;

		while (count-- > 0)
		{
			*dest = pal8_to24[*source];
			source++;
			dest += cachewidth;
		}

		column = (column_t *)((byte *)column + column->length + 4);
	}
}

//==========================================================================
//
//	TOpenGLDrawer::GenerateTexture
//
// 	Using the texture definition, the composite texture is created from the
// patches, and each column is cached.
//
//==========================================================================

void TOpenGLDrawer::GenerateTexture(int texnum)
{
	rgba_t*			block;
	texdef_t*		texture;
	texpatch_t*		patch;
	patch_t*		realpatch;
	int				x;
	int				x1;
	int				x2;
	int				i;
	column_t*		patchcol;
	int				htimes;
	int				texture_width;
	int				workh;

	texture = textures[texnum];

	texture_width = texture->width;
	if (texture_width > 256)
	{
		cond << "JL in trouble - texture " << texture->name
			<< " haves width " << texture_width << " greater than 256\n";
		texture_width = 256;
	}
	if (texture_width & (texture_width - 1))
	{
		//	This will happen only with Doom's dummy texture
		con << "Texture width is not a power of 2\n";
		texture_width = 16;
	}
	if (texture->height & (texture->height - 1))
	{
		//	Texture height is not a power of 2, use 256
		workh = 256;
	}
	else
	{
		workh = texture->height;
	}

	block = (rgba_t*)Z_Calloc(4 * texture_width * workh);

	// Composite the columns together.
	patch = texture->patches;

	htimes = workh / texture->height;

	for (i = 0; i < texture->patchcount; i++, patch++)
	{
		realpatch = (patch_t*)W_CacheLumpNum(patch->patch, PU_CACHE);
		x1 = patch->originx;
		x2 = x1 + LittleShort(realpatch->width);

		if (x1 < 0)
			x = 0;
		else
			x = x1;

		if (x2 > texture_width)
			x2 = texture_width;

		for ( ; x < x2; x++)
		{
			patchcol = (column_t *)((byte *)realpatch
					+ LittleLong(realpatch->columnofs[x - x1]));
			for (int ht = 0; ht < htimes; ht++)
			{
				DrawColumnInCache(patchcol, block +
					ht * texture->height * texture_width, x, patch->originy,
					texture_width, texture->height);
			}
		}
	}

	// Generate The Texture
	glTexImage2D(GL_TEXTURE_2D, 0, 4, texture_width, workh, 0, GL_RGBA, GL_UNSIGNED_BYTE, block);

	Z_Free(block);
	texture_sent[texnum] = true;
	texture_iw[texnum] = 1.0 / (float)texture_width;
	texture_ih[texnum] = 1.0 / (float)workh;
}

//==========================================================================
//
// 	TOpenGLDrawer::SetTexture
//
//==========================================================================

void TOpenGLDrawer::SetTexture(int tex)
{
	tex = texturetranslation[tex];

	glBindTexture(GL_TEXTURE_2D, texture_id[tex]);
	if (!texture_sent[tex])
	{
		GenerateTexture(tex);
	}
	if (tex_linear)
	{
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
	else
	{
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	}
	tex_iw = texture_iw[tex];
	tex_ih = texture_ih[tex];
}

//==========================================================================
//
//	GenerateSkyTexture
//
//==========================================================================

#if 0
static void GenerateSkyTexture(int texnum)
{
	byte*			block;
	texdef_t*		texture;
	texpatch_t*		patch;
	patch_t*		realpatch;
	int				x;
    int				x1;
    int				x2;
    int				i;
	column_t*		patchcol;
	int				wtimes;
	int				htimes;

    texture = textures[texnum];

	block = (byte*)Z_Malloc(1024 * 256, PU_STATIC, (void**)&skytexturedata[texnum]);
	memset(block, TRANSPARENT_COLOR, 1024 * 256);

    // Composite the columns together.
	patch = texture->patches;

	wtimes = 1024 / texture->width;
	htimes = 256 / texture->height;

    for (i = 0; i < texture->patchcount; i++, patch++)
    {
		realpatch = (patch_t*)W_CacheLumpNum(patch->patch, PU_CACHE);
		x1 = patch->originx;
		x2 = x1 + LittleShort(realpatch->width);

		if (x1 < 0)
	    	x = 0;
		else
	    	x = x1;
	
		if (x2 > texture->width)
			x2 = texture->width;

		for ( ; x < x2; x++)
		{
	    	patchcol = (column_t *)((byte *)realpatch
					+ LittleLong(realpatch->columnofs[x - x1]));
			for (int ht = 0; ht < htimes; ht++)
			{
				for (int wt = 0; wt < wtimes; wt++)
				{
			    	DrawColumnInCache(patchcol, block + wt * texture->width +
			    		ht * texture->height * 1024, x, patch->originy,
						1024, texture->height);
				}
			}
		}
    }

    // Now that the texture has been built in column cache,
	//  it is purgable from zone memory.
	Z_ChangeTag(block, PU_CACHE);
}
#endif

//==========================================================================
//
// 	TOpenGLDrawer::SetSkyTexture
//
//==========================================================================

void TOpenGLDrawer::SetSkyTexture(int tex, bool double_sky)
{
	rgba_t saved;
	if (double_sky)
	{
		saved = pal8_to24[0];
		pal8_to24[0].a = 0;
	}
#if 0
	tex = texturetranslation[tex];

    if (!skytexturedata[tex])
		GenerateSkyTexture(tex);

    ds_source = skytexturedata[tex];
#else
	SetTexture(tex);
#endif
	if (double_sky)
	{
		pal8_to24[0] = saved;
	}
}

//==========================================================================
//
//	TOpenGLDrawer::GenerateFlat
//
//==========================================================================

void TOpenGLDrawer::GenerateFlat(int num)
{
	rgba_t *block = (rgba_t*)Z_Malloc(4 * 64 * 64);
	byte *data = (byte*)W_CacheLumpNum(flatlumps[num], PU_CACHE);
	for (int i = 0; i < 64 * 64; i++)
	{
		block[i] = pal8_to24[data[i]];
	}

	// Generate The Texture
	glTexImage2D(GL_TEXTURE_2D, 0, 4, 64, 64, 0, GL_RGBA, GL_UNSIGNED_BYTE, block);

	Z_Free(block);
	flat_sent[num] = true;
}

//==========================================================================
//
//	TOpenGLDrawer::SetFlat
//
//==========================================================================

void TOpenGLDrawer::SetFlat(int num)
{
	num &= ~TEXF_FLAT;
	num = flattranslation[num];

	glBindTexture(GL_TEXTURE_2D, flat_id[num]);

	if (!flat_sent[num])
	{
		GenerateFlat(num);
	}
	if (tex_linear)
	{
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
	else
	{
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	}
	tex_iw = 1.0 / 64.0;
	tex_ih = 1.0 / 64.0;
}

//==========================================================================
//
//	TOpenGLDrawer::GenerateSprite
//
//==========================================================================

void TOpenGLDrawer::GenerateSprite(int lump)
{
    patch_t	*patch = (patch_t*)W_CacheLumpNum(spritelumps[lump], PU_STATIC);

	int w = LittleShort(patch->width);
	int h = LittleShort(patch->height);
	int p2w = ToPowerOf2(w);
	int p2h = ToPowerOf2(h);
	fixed_t xscale;
	if (p2w > maxTexSize)
	{
		xscale = FRACUNIT * w / maxTexSize;
		p2w = maxTexSize;
		spriteiw[lump] = 1.0 / (float)w;
	}
	else
	{
		xscale = FRACUNIT;
		spriteiw[lump] = 1.0 / (float)p2w;
	}
	spriteih[lump] = 1.0 / (float)p2h;
	int workw = MIN(w, p2w);

    rgba_t *block = (rgba_t*)Z_Calloc(4 * p2w * p2h);

	for (int x = 0; x < workw; x++)
	{
    	column_t *column = (column_t *)((byte *)patch +
    		LittleLong(patch->columnofs[(x * xscale) >> FRACBITS]));

		// step through the posts in a column
	    while (column->topdelta != 0xff)
	    {
		    byte* source = (byte *)column + 3;
		    rgba_t* dest = block + x + column->topdelta * p2w;
			int count = column->length;

	    	while (count--)
	    	{
				*dest = pal8_to24[*source];
				source++;
				dest += p2w;
	    	}
			column = (column_t *)((byte *)column + column->length + 4);
	    }
	}

	// Generate The Texture
	glTexImage2D(GL_TEXTURE_2D, 0, 4, p2w, p2h, 0, GL_RGBA, GL_UNSIGNED_BYTE, block);
	sprite_sent[lump] = true;

	Z_Free(block);
	Z_ChangeTag(patch, PU_CACHE);
}

//==========================================================================
//
//	TOpenGLDrawer::GenerateTranslatedSprite
//
//==========================================================================

void TOpenGLDrawer::GenerateTranslatedSprite(int lump, int slot, int translation)
{
    patch_t	*patch = (patch_t*)W_CacheLumpNum(spritelumps[lump], PU_STATIC);

	int w = LittleShort(patch->width);
	int h = LittleShort(patch->height);
	int p2w = ToPowerOf2(w);
	int p2h = ToPowerOf2(h);
	fixed_t xscale;
	if (p2w > maxTexSize)
	{
		xscale = FRACUNIT * w / maxTexSize;
		p2w = maxTexSize;
		trspriw[lump] = 1.0 / (float)w;
	}
	else
	{
		xscale = FRACUNIT;
		trspriw[lump] = 1.0 / (float)p2w;
	}
	trsprih[lump] = 1.0 / (float)p2h;
	int workw = MIN(w, p2w);

    rgba_t *block = (rgba_t*)Z_Calloc(4 * p2w * p2h);
	trspr_lump[slot] = lump;
	trspr_tnum[slot] = translation;
	trspr_sent[slot] = true;

	byte *trtab = translationtables + translation * 256;

	for (int x = 0; x < workw; x++)
	{
    	column_t *column = (column_t *)((byte *)patch +
    		LittleLong(patch->columnofs[(x * xscale) >> FRACBITS]));

		// step through the posts in a column
	    while (column->topdelta != 0xff)
	    {
		    byte* source = (byte *)column + 3;
		    rgba_t* dest = block + x + column->topdelta * p2w;
			int count = column->length;

	    	while (count--)
	    	{
				*dest = pal8_to24[trtab[*source]];
				source++;
				dest += p2w;
	    	}
			column = (column_t *)((byte *)column + column->length + 4);
	    }
	}

	// Generate The Texture
	glTexImage2D(GL_TEXTURE_2D, 0, 4, p2w, p2h, 0, GL_RGBA, GL_UNSIGNED_BYTE, block);

	Z_Free(block);
	Z_ChangeTag(patch, PU_CACHE);
}

//==========================================================================
//
//	TOpenGLDrawer::SetSpriteLump
//
//==========================================================================

void TOpenGLDrawer::SetSpriteLump(int lump, int translation)
{
	if (translation)
	{
		int i;
		int avail = -1;
		for (i = 0; i <	MAX_TRANSLATED_SPRITES; i++)
		{
			if (trspr_sent[i])
			{
				if (trspr_lump[i] == lump && trspr_tnum[i] == translation)
				{
					glBindTexture(GL_TEXTURE_2D, trspr_id[i]);
					if (tex_linear)
					{
						glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
						glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
					}
					else
					{
						glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
						glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
					}
					tex_iw = trspriw[i];
					tex_ih = trsprih[i];
					return;
				}
			}
			else
			{
				if (avail < 0)
					avail = i;
			}
		}
		if (avail < 0)
		{
			avail = 0;
		}
		glBindTexture(GL_TEXTURE_2D, trspr_id[avail]);
		GenerateTranslatedSprite(lump, avail, translation);
		if (tex_linear)
		{
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		}
		else
		{
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		}
		tex_iw = trspriw[avail];
		tex_ih = trsprih[avail];
	}
	else
	{
		glBindTexture(GL_TEXTURE_2D, sprite_id[lump]);
		if (!sprite_sent[lump])
		{
			GenerateSprite(lump);
		}
		if (tex_linear)
		{
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		}
		else
		{
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		}
		tex_iw = spriteiw[lump];
		tex_ih = spriteih[lump];
	}
}

//==========================================================================
//
//	TOpenGLDrawer::SetPic
//
//==========================================================================

void TOpenGLDrawer::SetPic(int handle)
{
	glBindTexture(GL_TEXTURE_2D, pic_id[handle]);

	if (!pic_sent[handle])
	{
		switch (pic_list[handle].type)
 		{
	 	 case PIC_PATCH:
			GeneratePicFromPatch(handle);
			break;

		 case PIC_RAW:
			GeneratePicFromRaw(handle);
			break;
		}
	}

	tex_iw = pic_iw[handle];
	tex_ih = pic_ih[handle];

	if (tex_linear)
	{
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
	else
	{
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	}
}

//==========================================================================
//
//	TOpenGLDrawer::GeneratePicFromPatch
//
//==========================================================================

void TOpenGLDrawer::GeneratePicFromPatch(int handle)
{
	patch_t *patch = (patch_t*)W_CacheLumpName(pic_list[handle].name, PU_STATIC);
	int w = LittleShort(patch->width);
	int h = LittleShort(patch->height);
	int p2w = ToPowerOf2(w);
	int p2h = ToPowerOf2(h);
	fixed_t xscale;
	if (p2w > maxTexSize)
	{
		xscale = FRACUNIT * w / maxTexSize;
		p2w = maxTexSize;
		pic_iw[handle] = 1.0 / (float)w;
	}
	else
	{
		xscale = FRACUNIT;
		pic_iw[handle] = 1.0 / (float)p2w;
	}
	pic_ih[handle] = 1.0 / (float)p2h;
	int workw = MIN(w, p2w);

    rgba_t *block = (rgba_t*)Z_Calloc(4 * p2w * p2h);

	for (int x = 0; x < workw; x++)
	{
    	column_t *column = (column_t *)((byte *)patch +
    		LittleLong(patch->columnofs[(x * xscale) >> FRACBITS]));

		// step through the posts in a column
	    while (column->topdelta != 0xff)
	    {
		    byte* source = (byte *)column + 3;
		    rgba_t* dest = block + x + column->topdelta * p2w;
			int count = column->length;

	    	while (count--)
	    	{
				*dest = pal8_to24[*source];
				source++;
				dest += p2w;
	    	}
			column = (column_t *)((byte *)column + column->length + 4);
	    }
	}

	glTexImage2D(GL_TEXTURE_2D, 0, 4, p2w, p2h, 0, GL_RGBA, GL_UNSIGNED_BYTE, block);
	pic_sent[handle] = true;
	Z_Free(block);
	Z_ChangeTag(patch, PU_CACHE);
}

//==========================================================================
//
//	TOpenGLDrawer::GeneratePicFromRaw
//
//==========================================================================

void TOpenGLDrawer::GeneratePicFromRaw(int handle)
{
	int p2w = 512;
	int p2h = 256;
	fixed_t xscale;
	if (p2w > maxTexSize)
	{
		p2w = maxTexSize;
		xscale = FRACUNIT * 320 / maxTexSize;
		pic_iw[handle] = 1.0 / 320.0;
	}
	else
	{
		xscale = FRACUNIT;
		pic_iw[handle] = 1.0 / 512.0;
	}
	pic_ih[handle] = 1.0 / 256.0;
	int workw = MIN(320, p2w);

	byte *raw = (byte*)W_CacheLumpName(pic_list[handle].name, PU_STATIC);
    rgba_t *block = (rgba_t*)Z_Calloc(4 * p2w * p2h);

	for (int y = 0; y < 200; y++)
	{
		byte *src = raw + 320 * y;
		rgba_t *dst = block + p2w * y;
		for (int x = 0; x < workw; x++)
		{
			dst[x] = pal8_to24[src[(x * xscale) >> FRACBITS]];
		}
	}

	glTexImage2D(GL_TEXTURE_2D, 0, 4, p2w, p2h, 0, GL_RGBA, GL_UNSIGNED_BYTE, block);
	Z_Free(block);
	Z_ChangeTag(raw, PU_CACHE);

	pic_sent[handle] = true;
}


